# -*- coding: utf-8 -*-
import sys
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.output import color_print
from pawnlib.config import pawn
from pawnlib.utils import http
from pawnlib.typing import date_utils, str2bool


__version__ = "0.0.1"
__description__ = "Connect to the Goloop network with WebSocket to receive blocks."


def get_parser():
    parser = argparse.ArgumentParser(description='Websocket')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument('url', help='url')
    parser.add_argument('-c', '--command', type=str, help='command', default=None, choices=["start", "stop", "restart", None])
    parser.add_argument('-v', '--verbose', action='count', help='verbose mode. view level', default=0)
    parser.add_argument('-b', '--blockheight', type=int, help='position of blockheight to start. ', default=0)
    parser.add_argument('-d', '--diff', type=int, help='diff timestamp ', default=2)
    parser.add_argument('-t', '--target', type=str, nargs='+', help='Monitoring target (block|tx)',
                        default=["block", "tx"], choices=["block", "tx"])
    parser.add_argument('--stack-limit', type=int, help='Stack limit count for notify', default=3)
    parser.add_argument('--wait', type=str2bool, help='waiting for response', default=1)
    return parser


def main():
    banner = generate_banner(
        app_name="websocket",
        author="jinwoo",
        description="Connect to Goloop with WebSocket",
        font="graffiti",
        version=_version
    )
    print(banner)
    # pawn_http.disable_ssl_warnings()
    parser = get_parser()
    args, unknown = parser.parse_known_args()

    pawn.console.log(args)

    args.try_pass = False
    pawn.set(
        args=args,
        try_pass=False,
        LAST_EXECUTE_POINT=0,
    )

    if args.verbose > 2:
        pawn.set(PAWN_DEBUG=True)

    http.GoloopWebsocket(
        connect_url=args.url,
        blockheight=args.blockheight,
        monitoring_target=args.target,
        verbose=args.verbose,
        sec_thresholds=args.diff

    ).run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import socket
import select
import time
import sys
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawnlib_config as pawn
from pawnlib.utils.http import remove_http, ALLOWS_HTTP_METHOD
from pawnlib.typing.check import is_int, is_valid_ipv4, detect_encoding
from collections import OrderedDict
import re
import ssl


def get_parser():
    parser = argparse.ArgumentParser(description='Proxy Reflector')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument("--listen", "-l", type=str, help="Listen  IPaddr:port", default="0.0.0.0:8080")
    parser.add_argument("--forward", "-f", type=str, help="Forward IPaddr:port", default=None, required=True)
    parser.add_argument("--buffer-size", type=int, help="buffer size for socket", default=4096)
    parser.add_argument("--delay", type=float, help="buffer delay for socket", default=0.0001)
    parser.add_argument("--timeout", '-t', type=float, help="timeout for socket", default=3)
    return parser


class Forward:
    def __init__(self):
        self._forwarder = None
        self._secure_forwarder = None

    def start(self, host, port, is_ssl=False, hostname="", timeout=10):
        self._forwarder = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._forwarder.connect((host, port))
            self._forwarder.settimeout(timeout)

            if is_ssl:
                context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
                self._secure_forwarder = context.wrap_socket(self._forwarder, server_hostname=hostname)
                return self._secure_forwarder

            return self._forwarder

        except Exception as e:
            pawn.console.log(f"[red][Forward ERROR] Connect to {host}:{port} {e}")
            return False


class EchoWebServer:
    input_list = []
    channel = {}

    def __init__(self, listen, forward, buffer_size, delay):
        self.data = None
        self.s = None

        self.listen = listen
        self.forward = forward

        self.l_host = None
        self.l_port = None
        self.f_host = None
        self.f_port = None

        self.is_ssl = False

        self.buffer_size = buffer_size
        self.delay = delay

        self._parse_ip_port()
        self._validate()
        self._listen_server()

        self._http_headers = {}
        self._ongoing_response = True

        self._request = {}

    def _parse_ip_port(self):
        _parsed_listen = parse_ip_port(self.listen)
        self.l_host = _parsed_listen.get('ip_or_domain')
        self.l_port = _parsed_listen.get('port')
        _parsed_forward = parse_ip_port(self.forward)
        self.f_host = _parsed_forward.get('ip_or_domain')
        self.f_port = _parsed_forward.get('port')
        self.f_hostname = _parsed_forward.get('hostname', '')
        self.is_ssl = _parsed_forward.get('is_ssl', '')
        pawn.console.debug(f"forward={_parsed_forward}")
        pawn.console.log(f"Listen {self.l_host}:{self.l_port} => Forward {self.f_hostname}({self.f_host}:{self.f_port})")

    def _listen_server(self):
        try:
            self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server.bind((self.l_host, self.l_port))
            self.server.listen(200)
            pawn.console.log(f"Server started on {self.l_host}:{self.l_port}")
        except Exception as e:
            pawn.console.log(f"Error starting server: {e}")
            raise

    def _validate(self):
        for ipaddr in [self.l_host, self.f_host]:
            if not is_valid_ipv4(ipaddr):
                # pawn.console.print(f"[bold red]Invalid IP address - {ipaddr}, {is_valid_ipv4(ipaddr)}")
                raise ValueError(f"Invalid IP address - '{ipaddr}'")

    def main_loop(self):
        self.input_list.append(self.server)
        while True:
            time.sleep(self.delay)
            ss = select.select
            input_ready, output_ready, except_ready = ss(self.input_list, [], [])
            for self.s in input_ready:
                if self.s == self.server:
                    self.on_accept()
                    break
                try:
                    self.data = self.s.recv(self.buffer_size)
                except Exception as e:
                    pawn.console.log(f"[red bold] {e} data={self.data}")
                if len(self.data) == 0:
                    self.on_close()
                    break
                else:
                    self.on_recv()

    def _parse_headers(self):
        self._http_headers = {}
        lines = self._decode_data()
        if lines and "\r\n" in lines:
            lines_arr = lines.split('\r\n')
            if len(lines_arr) >= 1:
                for line in lines_arr[1:]:
                    if line and ":" in line:
                        header_kv = line.split(":")
                        header_key = header_kv[0]
                        header_value = header_kv[1]
                        self._http_headers[header_key.strip()] = header_value.strip()

    def _decode_data(self, default_encode="utf8"):
        lines = ""
        try:
            encoding = detect_encoding(self.data, default_encode=default_encode)
            pawn.console.debug(f"encoding={encoding}")
            lines = self.data.decode(encoding, "ignore")
        except Exception as e:
            pawn.console.log(f"[red] [Exception] {e} data={self.data}")

        return lines

    def _get_decoded_data_lines(self, delimiter="\r\n") -> list:
        lines = self._decode_data()
        data_arr = lines.split(delimiter)
        return data_arr

    def _inject_headers(self, headers={}):
        self._parse_headers()
        lines = self._decode_data()
        # if lines and self.f_hostname and self._http_headers.get('Host'):
        if lines and self.f_hostname and self._http_headers.get('Host'):

            request_line, _headers = lines.split('\r\n', 1)
            if not request_line.startswith('HTTP'):
                pawn.console.log("START injection")
                headers_list =_headers.split('\r\n')[:-2]  # exclude the last two elements as they are empty
                headers_dict = OrderedDict((header.split(':')[0], ':'.join(header.split(':')[1:]).strip()) for header in headers_list)

                for header_k, header_v in headers.items():
                    if header_k and header_v:
                        headers_dict[header_k] = header_v
                headers_string = '\r\n'.join(f'{k}: {v}' for k, v in headers_dict.items())
                # Add the request line and the final two '\r\n' back in
                headers = f'{request_line}\r\n{headers_string}\r\n\r\n'
                self.data = headers.encode()

    def _parse_response(self):
        pawn.console.log(f"raw_data = {self.data}")
        lines = self._decode_data()
        pawn.console.log(f"lines={lines}")
        if lines:
            _request_line, _headers_body = lines.split('\r\n', 1)
            _headers, _body = _headers_body.split('\r\n\r\n')
            pawn.console.log(f"request_line={_request_line}")
            pawn.console.log(f"headers={_headers}")
            pawn.console.log(f"body={_body}")

    def parse_http_request(self):
        headers = {}
        # Split the request into lines
        request = self._decode_data()
        lines, body = request.split("\r\n\r\n")
        _headers = lines.split('\r\n')
        if _headers:
            self._request = {}
            # The first line contains the request method, path, and protocol version
            method, path, protocol = _headers[0].split(' ')
            # Parse the headers
            for header in _headers[1:]:
                if header and ": " in header:
                    key, value = header.split(': ')
                    if key and value:
                        headers[key.strip()] = value.strip()

            self._request = {
                'method': method,
                'path': path,
                'protocol': protocol,
                'headers': headers,
                'body': body
            }

    def parse_http_response(self):
        pawn.console.log(self.data)

    def convert_to_request_raw(self):
        _raw = ""
        if self._request:
            _raw = f"{self._request.get('method')} {self._request.get('path')} {self._request.get('protocol')}\r\n"
            if self._request.get('headers'):
                for k, v in self._request['headers'].items():
                    _raw += f"{k}: {v}\r\n"
            _raw += f"\r\n{self._request.get('body')}"
        self.data = _raw.encode()

    def modify_http_request_headers(self, headers={}):
        if self._request and self._request.get('headers') and headers:
            for k, v in headers.items():
                self._request['headers'][k] = v
                pawn.console.debug(f"[yellow] modified header => {k}: {v}")
            self.convert_to_request_raw()

    def on_accept(self):
        try:
            forward = Forward().start(self.f_host, self.f_port, is_ssl=self.is_ssl, hostname=self.f_hostname)
            pawn.console.log(f"[bold red]on_accept [/bold red] forward={forward}")
            clientsock, client_addr = self.server.accept()
            client_ip, client_port = client_addr
            if forward:
                pawn.console.rule(f"{client_ip}:{client_port} has connected", style="spring_green2", align="left")

                self.input_list.append(clientsock)
                self.input_list.append(forward)
                self.channel[clientsock] = forward
                self.channel[forward] = clientsock
            else:
                pawn.console.log(f"Can't establish connection with remote server. "
                                 f"Closing connection with client side: {client_ip}:{client_port}")
                clientsock.close()
        except Exception as e:
            pawn.console.log(f"Error handling connection: {e}")
            raise

    def on_close(self):
        try:
            self._ongoing_response = False
            # pawn.console.log(f"{self._return_ip_port(self.s.getpeername())}", "has disconnected")
            pawn.console.rule(f"{self._return_ip_port(self.s.getpeername())} has disconnected", style="spring_green2", align="right")

            # remove objects from input_list
            self.input_list.remove(self.s)
            self.input_list.remove(self.channel[self.s])
            out = self.channel[self.s]
            # close the connection with client
            self.channel[out].close()  # equivalent to do self.s.close()
            # close the connection with remote server
            self.channel[self.s].close()
            # delete both objects from channel dict
            del self.channel[out]
            del self.channel[self.s]
        except Exception as e:
            pawn.console.log(f"Error closing connection: {e}")
            raise

    @staticmethod
    def _is_http_method(data):
        for method in ALLOWS_HTTP_METHOD:
            if data.startswith(method.upper()):
                return True
        return False

    def on_recv(self):
        lines = self._decode_data()
        if lines and "\r\n" in lines:
            data_arr = self._get_decoded_data_lines()

            if data_arr and data_arr[0].startswith("HTTP"):
                if self.f_hostname:
                    _forward_host = f"{self.f_hostname}({self.f_host})"
                else:
                    _forward_host = f"{self.f_host}"
                pawn.console.log(f"[bold blue]Response from {_forward_host}:{self.f_port}")
                pawn.console.rule(f"[bold blue] ", align="right")

            elif data_arr and self._is_http_method(data_arr[0]):
                pawn.console.log(f"[bold blue]Request from {self.l_host}:{self.l_port}")
                pawn.console.debug(f"[green] request-> {self.data}")
                self.parse_http_request()
                self.modify_http_request_headers(headers={'Host': self.f_hostname})
                pawn.console.debug(f"[bold blue] {self.data}")
                data_arr = self._get_decoded_data_lines()

            elif self._ongoing_response:
                pawn.console.log(f"[green] request-> {self.data}")
                pawn.console.log(f"[bold blue]Request from {self.l_host}:{self.l_port}")
                pawn.console.rule(f"[bold blue] ", align="right")
                self._ongoing_response = False

            for line in data_arr:
                pawn.console.print(line)

        self.channel[self.s].send(self.data)

    @staticmethod
    def _return_ip_port(data):
        if data and len(data) > 1:
            return "{}:{}".format(*data)


def resolve_domain_to_ip(domain):
    try:
        ip = socket.gethostbyname(domain)
        return ip
    except socket.gaierror:
        return None


def is_valid_ip(ip):
    # Regular expression pattern for IP address validation
    pattern = r'^((?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))$'

    # Match IP address using regex
    match = re.match(pattern, ip)
    return match is not None


def parse_ip_port(input_str):
    # Regular expression pattern to match IP address, domain name, and port
    pattern = r'(?:(?:http|https)://)?(?:([\w_-]+(?:\.[\w_-]+)*|\d{1,3}(?:\.\d{1,3}){3}))(?::(\d+))?(?:/|$)'
    match = re.search(pattern, input_str)

    if match:
        ip_or_domain = match.group(1)
        port = match.group(2)

        return dict(
            ip_or_domain=resolve_domain_to_ip(ip_or_domain),
            port=int(port) if port else 443 if 'https' in input_str else 80,
            hostname=ip_or_domain if not is_valid_ip(ip_or_domain) else "",
            is_ssl=True if input_str.startswith("https://") else False
        )
    return {}


def main():

    banner = generate_banner(
        app_name="proxy reflector",
        author="jinwoo",
        description="proxy reflector",
        font="graffiti",
        version=_version
    )
    print(banner)

    parser = get_parser()
    args, unknown = parser.parse_known_args()

    pawn.console.log(f"args = {args}")

    server = EchoWebServer(
        listen=args.listen,
        forward=args.forward,
        buffer_size=args.buffer_size,
        delay=args.delay,
    )
    try:
        server.main_loop()
    except KeyboardInterrupt:
        print("Ctrl C - Stopping server")
        sys.exit(1)


if __name__ == '__main__':
    main()




import inspect
import multiprocessing
import os
import sys
import json
# import sqlite3


class FirstRunCheckerSqlite:

    _instance = None
    _lock = multiprocessing.Lock()

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, db_file=".task_first_run.db", debug=False):
        if hasattr(self, "initialized"):
            return
        self.initialized = True

        self.db_file = db_file
        self.key = "default"
        self.debug = debug
        self.setup_database()
        # self.debug_print("Initialized")

    def setup_database(self):
        conn = sqlite3.connect(self.db_file)
        cursor = conn.cursor()
        cursor.execute('CREATE TABLE IF NOT EXISTS first_run_flags (key TEXT PRIMARY KEY, flag INTEGER)')
        conn.commit()
        conn.close()

    def is_first_run(self, key=None):
        if key:
            self.key = key
        conn = sqlite3.connect(self.db_file)
        cursor = conn.cursor()
        cursor.execute('SELECT flag FROM first_run_flags WHERE key=?', (self.key,))
        result = cursor.fetchone()
        conn.close()
        return not result or not bool(result[0])

    def one_time_run(self, key=None):
        if key:
            self.key = key
        else:
            caller_frame = inspect.currentframe().f_back
            caller_module = inspect.getmodule(caller_frame)
            key = f"{caller_module.__name__}.{caller_frame.f_code.co_name}"  # Add module name to the key
            self.key = key

        with self._lock:
            _is_first_run_result = self.is_first_run(self.key)
            if _is_first_run_result:
                self.mark_first_run(self.key)
                return True
            return False

    def mark_first_run(self, key=None):
        if key:
            self.key = key
        conn = sqlite3.connect(self.db_file)
        cursor = conn.cursor()
        cursor.execute('INSERT OR REPLACE INTO first_run_flags (key, flag) VALUES (?, 1)', (self.key,))
        conn.commit()
        conn.close()

    def clear_first_run(self):
        conn = sqlite3.connect(self.db_file)
        cursor = conn.cursor()
        cursor.execute('DELETE FROM first_run_flags')
        conn.commit()
        conn.close()


class FirstRunChecker:
    """
    A class to check if it's the first run of the program.

    This class uses a file to keep track of whether it's the first run of the program or not.
    The file path can be specified, and by default, it's ".task_first_run.json".

    :param file_path: The file path to keep track of the first run. Default is ".task_first_run.json".
    :param debug: A flag to print debug information. Default is False.

    Example:

        .. code-block:: python
            from pawnlib.config import FirstRunChecker, one_time_run
            checker = FirstRunChecker(file_path=".first_run.json", debug=True)
            if checker.is_first_run():
                print("This is the first run.")
            else:
                print("This is not the first run.")
            # >> This is the first run.

            checker.mark_first_run()
            if checker.is_first_run():
                print("This is the first run.")
            else:
                print("This is not the first run.")
            # >> This is not the first run.

            checker.clear_first_run()
            if checker.is_first_run():
                print("This is the first run.")
            else:
                print("This is not the first run.")
            # >> This is the first run.

            if one_time_run():
                print("one time run")
            #>> one time run

    """
    _instance = None
    _lock = multiprocessing.Lock()

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, file_path=".task_first_run.json", debug=False):
        if hasattr(self, "initialized"):
            return
        self.initialized = True

        self.file_path = file_path
        self.key = "default"
        self.debug = debug

        self._is_check_python_version = False
        self.data = self.load_data()

    def check_python_version(self):
        if not self._is_check_python_version:
            if sys.version_info.major != 3 or sys.version_info.minor != 7:
                current_version = f"Python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
                print(f"Warning: This function can only be used with Python 3.7. Your current version is {current_version}")
                self._is_check_python_version = True

    def load_data(self):
        if os.path.exists(self.file_path):
            with open(self.file_path, "r") as file:
                try:
                    data = json.load(file)
                except json.JSONDecodeError:
                    data = {}
        else:
            data = {}
        return data

    def sync_data(self):
        self.data = self.load_data()

    def save_data(self):
        with open(self.file_path, "w") as file:
            json.dump(self.data, file)

    def is_first_run(self, key=None):
        self.sync_data()
        if key:
            self.key = key
        return not self.data.get(self.key, False)

    def one_time_run(self, key=None):
        self.check_python_version()

        if key:
            self.key = key
        else:
            caller_frame = inspect.currentframe().f_back
            caller_module = inspect.getmodule(caller_frame)
            key = f"{caller_module.__name__}.{caller_frame.f_code.co_name}"  # Add module name to the key
            self.key = key

        with self._lock:
            _is_first_run_result = self.is_first_run(self.key)
            if _is_first_run_result:
                self.mark_first_run(self.key)
                return True
            return False

    def mark_first_run(self, key=None):
        if key:
            self.key = key
        self.data[self.key] = True
        self.save_data()

    def clear_first_run(self):
        self.data = {}
        if os.path.exists(self.file_path):
            os.remove(self.file_path)


first_run_checker = FirstRunChecker()
first_run_checker.clear_first_run()
one_time_run = first_run_checker.one_time_run




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import sys
import os
import json
import getpass
import traceback
import inspect
import executing
from contextlib import contextmanager, AbstractContextManager

from pawnlib.typing import converter, date_utils, list_to_oneline_string, const, is_include_list, remove_tags, remove_ascii_color_codes,timestamp_to_string, is_hex
from pawnlib.config import pawnlib_config as pawn, global_verbose
from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import Terminal256Formatter
from rich.syntax import Syntax
from rich.table import Table
from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn
from rich import print as rprint
from typing import Union, Callable
from datetime import datetime
import textwrap
from requests.structures import CaseInsensitiveDict


_ATTRIBUTES = dict(
    list(zip([
        'bold',
        'dark',
        '',
        'underline',
        'blink',
        '',
        'reverse',
        'concealed'
    ],
        list(range(1, 9))
    ))
)
del _ATTRIBUTES['']

_HIGHLIGHTS = dict(
    list(zip([
        'on_grey',
        'on_red',
        'on_green',
        'on_yellow',
        'on_blue',
        'on_magenta',
        'on_cyan',
        'on_white'
    ],
        list(range(40, 48))
    ))
)

_COLORS = dict(
    list(zip([
        'grey',
        'red',
        'green',
        'yellow',
        'blue',
        'magenta',
        'cyan',
        'white',
    ],
        list(range(30, 38))
    ))
)

_RESET = '\033[0m'


def colored(text, color=None, on_color=None, attrs=None):
    """Colorize text.

    Available text colors:
        red, green, yellow, blue, magenta, cyan, white.

    Available text highlights:
        on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.

    Available _ATTRIBUTES:
        bold, dark, underline, blink, reverse, concealed.

    Example:
        colored('Hello, World!', 'red', 'on_grey', ['blue', 'blink'])
        colored('Hello, World!', 'green')
    """
    if os.getenv('ANSI_COLORS_DISABLED') is None:
        fmt_str = '\033[%dm%s'
        if color is not None:
            text = fmt_str % (_COLORS[color], text)

        if on_color is not None:
            text = fmt_str % (_HIGHLIGHTS[on_color], text)

        if attrs is not None:
            for attr in attrs:
                if attr is not None:
                    text = fmt_str % (_ATTRIBUTES[attr], text)

        text += _RESET
    return text


def cprint(text, color=None, on_color=None, attrs=None, **kwargs):
    """Print colorize text.
    It accepts arguments of print function.

    :param text:
    :param color:
    :param on_color:
    :param attrs:
    :param kwargs:
    :return:

    Example:
        .. code-block:: python

            cprint("message", "red")  # >>  message

    """

    print((colored(text, color, on_color, attrs)), **kwargs)


class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    GREEN = '\033[32;40m'
    CYAN = '\033[96m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    ITALIC = '\033[1;3m'
    UNDERLINE = '\033[4m'
    WHITE = '\033[97m'
    DARK_GREY = '\033[38;5;243m'
    LIGHT_GREY = '\033[37m'


class PrintRichTable:
    """

    Print a table using a rich.table module.

    :param title: Title of table
    :param data: Data of table
    :param columns: Columns of table. Print only column parameter values.
    :param with_idx: Print the row count.
    :param call_value_func: The row value must be a string. If you want to perform other tasks, please add the function name.
    :param call_desc_func: Invoke a function that describes the value.
    :param display_output: Determines whether to print output (default: True)
    :param justify: Alignment value of the console. (default: left)


    Example:

        .. code-block:: python

            from pawnlib.output.color_print import PrintRichTable

            data = [
                {
                    "address":         "1x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08",
                    "value":         399999999999999966445568,
                },
                {
                    "address":         "2x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08",
                    "value":         399999999999999966445568,
                },
                {
                    "address":         "3x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08",
                    "value":         399999999999999966445568,
                }
            ]

            PrintRichTable(title="RichTable", data=data)


            #                                    RichTable
            # ┏━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━┓
            # ┃ idx ┃ address                                    ┃ value                    ┃
            # ┡━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━┩
            # │ 0   │ 1x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08 │ 399999999999999966445568 │
            # ├─────┼────────────────────────────────────────────┼──────────────────────────┤
            # │ 1   │ 2x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08 │ 399999999999999966445568 │
            # ├─────┼────────────────────────────────────────────┼──────────────────────────┤
            # │ 2   │ 3x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08 │ 399999999999999966445568 │
            # └─────┴────────────────────────────────────────────┴──────────────────────────┘
            #
            # PrintRichTable(title="RichTable", data=data, columns=["address"])
            #
            #                               RichTable
            # ┏━━━━━┳━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┓
            # ┃ idx ┃ address                                    ┃
            # ┡━━━━━╇━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┩
            # │ 0   │ 1x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08 │
            # ├─────┼────────────────────────────────────────────┤
            # │ 1   │ 2x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08 │
            # ├─────┼────────────────────────────────────────────┤
            # │ 2   │ 3x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08 │
            # └─────┴────────────────────────────────────────────┘

    """

    def __init__(
            self,
            title: str = "",
            data: Union[dict, list] = None,
            columns: list = None,
            remove_columns: list = None,
            with_idx: bool = True,
            call_value_func=str,
            call_desc_func=None,
            columns_options=None,
            display_output=True,
            justify="left",
            **kwargs
    ) -> None:

        if columns is None:
            columns = list()
        if data is None:
            data = dict()
        self.title = f"[bold cyan] {title}"
        self.table_options = kwargs
        # self.table = Table(title=self.title, **kwargs)
        self.table = None
        self.data = data
        self.table_data = []
        self.columns = columns
        self.remove_columns = remove_columns
        self.rows = []
        self.row_count = 0
        self.with_idx = with_idx
        self.call_value_func = call_value_func
        self.call_desc_func = call_desc_func
        self.display_output = display_output
        self.console_justify = justify

        _default_columns_option = dict(
            key=dict(
                justify="left",
            ),
            value=dict(
                justify="right",
            ),
            description=dict(
                justify="right",
            ),
        )
        self.columns_options = _default_columns_option
        if columns_options:
            self.columns_options.update(columns_options)

        self._check_columns_options()
        self._initialize_table()
        self._set_table_data()
        self._print_table()

    def _check_columns_options(self):
        allowed_columns = ["header_style", "footer_style", "style", "justify", "vertical", "overflow", "width", "min_width", "max_width", "ratio",
                           "no_wrap"]
        for column_name, column_values in self.columns_options.items():
            if isinstance(column_values, dict):
                for column_key, value in column_values.items():
                    if column_key not in allowed_columns:
                        raise ValueError(f"name='{column_name}', key='{column_key}' is not allowed column option, allowed: {allowed_columns}")

    def _initialize_table(self):
        if isinstance(self.data, dict):
            self.table_data = self.data
            if not self.table_options.get('show_header'):
                self.table_options['show_header'] = False
            if self.table_options.get('show_lines', "NOT_DEFINED") == "NOT_DEFINED":
                self.table_options['show_lines'] = False

        elif isinstance(self.data, list):
            self.table_data = self.data
        else:
            self.table_data = []

        self.table = Table(title=self.title, **self.table_options)

    # def _specify_columns
    def _is_showing_columns(self, item):
        if len(self.columns) == 0:
            return True
        if item in self.columns:
            return True
        return False

    def _draw_vertical_table(self):
        pawn.console.debug("Drawing vertical table")
        if self.with_idx:
            self.table.add_column("idx", **self.columns_options.get('idx', {}))
        self.table.add_column("key", **self.columns_options.get('key', {}))
        self.table.add_column("value", **self.columns_options.get('value', {}))
        if self.call_desc_func and callable(self.call_desc_func):
            self.table.add_column("description", **self.columns_options.get('description', {}))

        _count = 0
        row_dict = {}
        for item, value in self.table_data.items():
            if self._is_showing_columns(item):
                row_dict[item] = value
                if callable(self.call_value_func):
                    value = self.call_value_func(value)

                columns = [f"{item}", f"{value}"]
                if self.with_idx:
                    columns.insert(0, f"{_count}")
                if self.call_desc_func and callable(self.call_desc_func):
                    columns.append(self.call_desc_func(*columns, **row_dict))

                self.table.add_row(*columns)
                _count += 1

    def _draw_horizontal_table(self):
        pawn.console.debug("Drawing horizontal table")
        for item in self.table_data:
            if isinstance(item, dict):
                line_row = []
                row_dict = {}
                for column in self.columns:
                    if self.with_idx and column == "idx":
                        value = str(self.row_count)
                    elif column == "desc":
                        value = self.call_desc_func(*line_row, **row_dict)
                    else:
                        try:
                            value = self.call_value_func(item.get(column), **{column: item.get(column)})
                        except Exception:
                            value = self.call_value_func(item.get(column))
                    row_dict[column] = value
                    line_row.append(value)
                self.rows.append(line_row)
            else:
                self.rows.append([f"{self.row_count}", f"{item}"])
            self.row_count += 1

        for col in self.columns:
            self.table.add_column(col, **self.columns_options.get(col, {}))

    def _extract_columns(self):
        # if self.table_data and len(self.columns) == 0 and isinstance(self.table_data[0], dict):
        if self.table_data and len(self.columns) == 0:
            try:
                self.columns = list(self.table_data[0].keys())
            except Exception as e:
                self.columns = ["value"]
        if self.with_idx:
            self.columns.insert(0, "idx")
        if callable(self.call_desc_func):
            self.columns.append("desc")

        if self.columns and isinstance(self.remove_columns, list):
            for r_column in self.remove_columns:
                self.columns.remove(r_column)

    def _set_table_data(self):
        if isinstance(self.table_data, list):
            self._extract_columns()
            self._draw_horizontal_table()
        elif isinstance(self.table_data, dict):
            self._draw_vertical_table()

    def _print_table(self):
        for row in self.rows:
            self.table.add_row(*row)

        if self.display_output:
            if self.table.columns:
                pawn.console.print(self.table, justify=self.console_justify)
            else:
                pawn.console.print(f"{self.title} \n  [i]No data ... [/i]")
        else:
            return self.table

class TablePrinter(object):
    "Print a list of dicts as a table"

    def __init__(self, fmt=[], sep='  ', ul="-"):
        """

        :param fmt: list of tuple(heading, key, width)

                    heading: str, column label \n
                    key: dictionary key to value to print \n
                    width: int, column width in chars \n

        :param sep: string, separation between columns
        :param ul: string, character to underline column label, or None for no underlining

        Example:
            .. code-block:: python

                from pawnlib import output

                data = [
                    {
                        "address":         "1x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08",
                        "value":         399999999999999966445568,
                    },
                    {
                        "address":         "2x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08",
                        "value":         399999999999999966445568,
                    },
                    {
                        "address":         "3x038bd14d5ce28a4ac713c21e89f0e6ca5f107f08",
                        "value":         399999999999999966445568,
                    },

                ]
                fmt = [
                    ('address',       'address',          10),
                    ('value',       'value',          15)
                ]
                cprint("Print Table", "white")
                print(output.TablePrinter(fmt=fmt)(data))
                print(output.TablePrinter()(data))

                 address         value
                ----------  ---------------
                1x038bd14d  399999999999999
                2x038bd14d  399999999999999
                3x038bd14d  399999999999999


        """
        super(TablePrinter, self).__init__()
        self._params = {"sep": sep, "ul": ul}

        self.fmt = str(sep).join('{lb}{0}:{1}{rb}'.format(key, width, lb='{', rb='}') for heading, key, width in fmt)
        self.head = {key: heading for heading, key, width in fmt}
        self.ul = {key: str(ul) * width for heading, key, width in fmt} if ul else None
        self.width = {key: width for heading, key, width in fmt}
        self.data_column = []

    def row(self, data, head=False):
        if head:
            return self.fmt.format(**{k: get_bcolors(data.get(k), "WHITE", bold=True, width=w) for k, w in self.width.items()})
        else:
            return self.fmt.format(**{k: str(data.get(k, ''))[:w] for k, w in self.width.items()})

    def get_unique_columns(self):
        self.data_column = []
        for item in self.data:
            self.data_column = self.data_column + list(item.keys())

        self.data_column = list(set(self.data_column))
        self.data_column.sort()

    def __call__(self, data_list):
        if len(self.fmt) == 0:
            sep = self._params.get("sep")
            ul = self._params.get("ul")

            self.data = data_list
            self.get_unique_columns()

            fmt = self.data_column

            width = 12
            self.fmt = str(sep).join('{lb}{0}:{1}{rb}'.format(key, width, lb='{', rb='}') for key in fmt)
            self.width = {key: width for key in fmt}
            self.ul = {key: str(ul) * width for key in fmt} if ul else None
            self.head = {key: key for key in fmt}

        _r = self.row
        res = [_r(data) for data in data_list]
        res.insert(0, _r(self.head, head=True))
        if self.ul:
            res.insert(1, _r(self.ul))
        return '\n'.join(res)


def get_bcolors(text, color, bold=False, underline=False, width=None):
    """

    Returns the color from the bcolors object.

    :param text:
    :param color:
    :param bold:
    :param underline:
    :param width:
    :return:
    """
    if width and len(text) <= width:
        text = text.center(width, ' ')
    return_text = f"{getattr(bcolors, color)}{text}{bcolors.ENDC}"
    if bold:
        return_text = f"{bcolors.BOLD}{return_text}"
    if underline:
        return_text = f"{bcolors.UNDERLINE}{return_text}"
    return str(return_text)


def colored_input(message, password=False, color="WHITE"):
    input_message = get_bcolors(text=message, color=color, bold=True, underline=True) + " "
    if password:
        return getpass.getpass(input_message)
    return input(input_message)


def get_colorful_object(v):
    if type(v) == bool:
        value = f"{bcolors.ITALIC}"
        if v is True:
            value += f"{bcolors.GREEN}"
        else:
            value += f"{bcolors.FAIL}"
        value += f"{str(v)}{bcolors.ENDC}"
    elif type(v) == int or type(v) == float:
        value = f"{bcolors.CYAN}{str(v)}{bcolors.ENDC}"
    elif type(v) == str:
        value = f"{bcolors.WARNING}'{str(v)}'{bcolors.ENDC}"
    elif v is None:
        value = f"{bcolors.FAIL}{str(v)}{bcolors.ENDC}"
    else:
        value = f"{v}"
    return value


def dump(obj, nested_level=0, output=sys.stdout, hex_to_int=False, debug=True, _is_list=False, _last_key=None, is_compact=False):
    """
    Print a variable for debugging.

    :param obj:
    :param nested_level:
    :param output:
    :param hex_to_int:
    :param debug:
    :param _is_list:
    :param _last_key:
    :param is_compact:
    :return:
    """
    spacing = '   '
    def_spacing = '   '
    format_number = lambda n: n if n % 1 else int(n)

    ignore_keys = ["Hash"]

    if type(obj) == dict:
        if nested_level == 0 or _is_list:
            print('%s{' % (def_spacing + (nested_level) * spacing))
        else:
            print("{")

        for k, v in obj.items():
            if hasattr(v, '__iter__'):
                print(bcolors.OKGREEN + '%s%s: ' % (def_spacing + (nested_level + 1) * spacing, k) + bcolors.ENDC, end="")
                dump(v, nested_level + 1, output, hex_to_int, debug, _last_key=k, is_compact=is_compact)
            else:
                if debug:
                    v = f"{get_colorful_object(v)} {bcolors.HEADER} {str(type(v)):>20}{bcolors.ENDC}{bcolors.DARK_GREY} len={len(str(v))}{bcolors.ENDC}"
                print(bcolors.OKGREEN + '%s%s:' % (def_spacing + (nested_level + 1) * spacing, k) + bcolors.WARNING + ' %s ' % v + bcolors.ENDC,
                      file=output)
        print('%s}' % (def_spacing + nested_level * spacing), file=output)
    elif type(obj) == list:
        if is_compact:
            end = ' '
        else:
            end = '\n'

        print('%s[' % (def_spacing + (nested_level) * spacing), file=output, end=end)
        for v in obj:
            if hasattr(v, '__iter__'):
                dump(v, nested_level + 1, output, hex_to_int, debug, _is_list=True, is_compact=is_compact)
            else:
                if is_compact:
                    spacing = ""
                    def_spacing = ''
                    end = ', '
                else:
                    end = '\n'
                # print(bcolors.WARNING + '%s%s' % (def_spacing + (nested_level + 1) * spacing, get_colorful_object(v)) + bcolors.ENDC, file=output)
                print(bcolors.WARNING + '%s%s' % (def_spacing + (nested_level + 1) * spacing, get_colorful_object(v)) + bcolors.ENDC, file=output,
                      end=end)
        print('%s]' % (def_spacing + (nested_level) * spacing), file=output)
    else:
        if debug:
            converted_hex = ""

            if hex_to_int and converter.is_hex(obj) and not is_include_list(target=_last_key, include_list=ignore_keys, ignore_case=False):
                if _last_key == "timestamp":
                    t_value = round(int(obj, 16) / 1_000_000)
                    converted_str = f"(from {_last_key})"
                    converted_date = datetime.fromtimestamp(format_number(t_value)).strftime('%Y-%m-%d %H:%M:%S')
                    converted_hex = f"{converted_date} {converted_str}"

                elif isinstance(obj, str) and len(obj) >= 60:
                    pass
                else:
                    if len(obj) < 14:
                        TINT = 1
                        TINT_STR = ""
                    else:
                        TINT = const.TINT
                        TINT_STR = f"(from TINT) {_last_key}"

                    converted_float = format_number(round(int(obj, 16) / TINT, 4))
                    converted_hex = f"{converted_float:,} {TINT_STR}"

            obj = f"{get_colorful_object(obj)}  " \
                  f"{bcolors.ITALIC}{bcolors.LIGHT_GREY}{converted_hex}{bcolors.ENDC}" \
                  f"{bcolors.HEADER} {str(type(obj)):>20}{bcolors.ENDC}" \
                  f"{bcolors.DARK_GREY} len={len(str(obj))}{bcolors.ENDC}"
        print(bcolors.WARNING + '%s%s' % (def_spacing + nested_level * spacing, obj) + bcolors.ENDC)


# def dump(obj, nested_level=1, output=sys.stdout, hex_to_int=False, debug=True, _last_key=None, is_compact=False):
#     """
#     Print a variable for debugging.
#
#     :param obj:
#     :param nested_level:
#     :param output:
#     :param hex_to_int:
#     :param debug:
#     :param _last_key:
#     :param is_compact:
#     :return:
#     """
#     if isinstance(obj, (dict, list)):
#         _print_iterable(obj, nested_level, output, hex_to_int, debug, _last_key, is_compact)
#     else:
#         _print_value(obj, '   ' * nested_level, _last_key, hex_to_int, debug)
#
#
# def _print_iterable(obj, nested_level, output, hex_to_int, debug, _last_key, is_compact):
#     spacing = '   ' * nested_level
#     print(f"{spacing}{'{' if isinstance(obj, dict) else '['}", end=("" if is_compact and isinstance(obj, list) else "\n"))
#     for k, v in (obj.items() if isinstance(obj, dict) else enumerate(obj)):
#         if isinstance(obj, dict):
#             print(f"{spacing}   {bcolors.OKGREEN}{k}: {bcolors.ENDC}", end="")
#         if isinstance(v, (dict, list)):
#             dump(v, nested_level + 1, output, hex_to_int, debug, _last_key=(k if isinstance(obj, dict) else _last_key), is_compact=is_compact)
#         else:
#             _print_value(v, spacing + '   ', (k if isinstance(obj, dict) else _last_key), hex_to_int, debug)
#         if is_compact and isinstance(obj, list):
#             print(", ", end="")
#     print(f"{' ' if is_compact and isinstance(obj, list) else spacing}{'}' if isinstance(obj, dict) else ']'}", end=("" if is_compact and isinstance(obj, list) else "\n"), file=output)
#
#
# def _print_value(v, spacing, _last_key, hex_to_int, debug):
#     output_value = _process_value(v, _last_key, hex_to_int)
#     if debug:
#         output_value += f" {bcolors.HEADER} {str(type(v)):>20}{bcolors.ENDC}{bcolors.DARK_GREY} len={len(str(v))}{bcolors.ENDC}"
#     print(f"{bcolors.WARNING}{spacing}{output_value}{bcolors.ENDC}")
#
#
# def _process_value(value, _last_key=None, hex_to_int=False):
#     _convert_based_key_dict = {
#         "time_stamp": timestamp_to_string
#     }
#
#     if hex_to_int and is_hex(value) and _last_key is not None:
#         return f"'{value}' {_process_hex(value, _last_key)}"
#     elif _convert_based_key_dict.get(_last_key):
#         _function = _convert_based_key_dict.get(_last_key)
#         return f"{get_colorful_object(value)}\t{_function(value)} (from {_last_key})"
#     else:
#         return get_colorful_object(value)
#
#
# def _process_hex(obj, _last_key):
#     format_number = lambda n: n if n % 1 else int(n)
#     if _last_key == "timestamp":
#         t_value = round(int(obj, 16) / 1_000_000)
#         converted_date = datetime.fromtimestamp(format_number(t_value)).strftime('%Y-%m-%d %H:%M:%S')
#         return f" {bcolors.ITALIC}{bcolors.LIGHT_GREY}{converted_date} (from {_last_key}){bcolors.ENDC}"
#     elif isinstance(obj, str) and len(obj) < 60:
#         TINT = 1 if len(obj) < 14 else 10 ** 18
#         TINT_STR = "" if len(obj) < 14 else f"(from TINT) {_last_key}"
#         converted_float = format_number(round(int(obj, 16) / TINT, 4))
#         return f" {bcolors.ITALIC}{bcolors.LIGHT_GREY}{converted_float:,} {TINT_STR}{bcolors.ENDC}"
#     else:
#         return ""


def debug_print(text, color="green", on_color=None, attrs=None, view_time=True, **kwargs):
    """Print colorize text.

    It accepts arguments of print function.
    """
    module_name = ''
    stack = inspect.stack()
    parent_frame = stack[1][0]
    module = inspect.getmodule(parent_frame)
    if module:
        module_pieces = module.__name__.split('.')
        module_name = list_to_oneline_string(module_pieces)
    function_name = stack[1][3]
    full_module_name = f"{module_name}.{function_name}({stack[1].lineno})"

    module_text = ""
    time_text = ""
    try:
        module_text = get_bcolors(f"[{full_module_name:<25}]", "WARNING")
    except:
        pass

    if view_time:
        time_text = "[" + get_bcolors(f"{date_utils.todaydate('log')}", "WHITE") + "]"
    main_text = (colored(str(text), color, on_color, attrs))
    print(f"{time_text}{module_text} {main_text}", **kwargs)


def classdump(obj):
    """
    For debugging, Print the properties of the class are shown.
    :param obj:
    :return:
    """
    for attr in dir(obj):
        if hasattr(obj, attr):
            value = getattr(obj, attr)
            print(bcolors.OKGREEN + f"obj.{attr} = " + bcolors.WARNING + f"{value}" + bcolors.ENDC)


def kvPrint(key, value):
    """
    print the  {key: value} format.

    :param key:
    :param value:
    :return:
    """
    key_width = 9
    key_value = 3
    print(bcolors.OKGREEN + "{:>{key_width}} : ".format(key, key_width=key_width) + bcolors.ENDC, end="")
    print(bcolors.WARNING + "{:>{key_value}} ".format(str(value), key_value=key_value) + bcolors.ENDC)


def print_json(obj, syntax=True, line_indent="", style="material",  **kwargs):
    """
    Print a JSON object with optional syntax highlighting and indentation.

    :param obj: The JSON object to print.
    :param syntax: Whether to use syntax highlighting (default: True).
    :param line_indent: The indentation for each line (default: "").
    :param style: Style for syntax highlighting (default: "material")
    :param kwargs: Additional keyword arguments for json.dumps().

    Example:

        .. code-block:: python

            data = {
                "name": "John",
                "age": 30,
                "city": "New York"
            }

            print_json(data)
            # >> {
            # >>     "name": "John",
            # >>     "age": 30,
            # >>     "city": "New York"
            # >> }

            print_json(data, syntax=False)
            # >> {"name": "John", "age": 30, "city": "New York"}

    """
    print(pretty_json(obj, syntax=syntax, line_indent=line_indent, style=style, **kwargs))


def pretty_json(obj, syntax=True, rich_syntax=False, style="one-dark", line_indent="", **kwargs):
    """
      Return a prettified JSON string with optional syntax highlighting.

      :param obj: JSON object to prettify
      :param syntax: If True, apply syntax highlighting (default: True)
      :param rich_syntax: If True, use rich library for syntax highlighting (default: False)
      :param style: Style name for syntax highlighting (default: "one-dark")
      :param line_indent: Custom line indentation (default: "")
      :param kwargs: Additional keyword arguments for json.dumps

      Example:

          .. code-block:: python

              data = {"name": "John", "age": 30, "city": "New York"}
              print(pretty_json(data))
              # {
              #     "name": "John",
              #     "age": 30,
              #     "city": "New York"
              # }

              print(pretty_json(data, syntax=False))
              # {"name": "John", "age": 30, "city": "New York"}

      """

    if syntax and isinstance(kwargs, dict):
        kwargs.setdefault("indent", 4)
        line_indent = " " * 4 if not line_indent else line_indent

    def json_to_string(_obj):
        if isinstance(_obj, str):
            _obj = json.loads(_obj)

        if isinstance(_obj, (dict, list)):
            return json.dumps(_obj, **kwargs)
        else:
            return _obj

    json_string = json_to_string(obj).strip()

    if syntax:
        if rich_syntax:
            return Syntax(json_string, "json", line_numbers=False, theme=style)
        else:
            return syntax_highlight(json_string, name="json", line_indent=line_indent, style=style).strip()
    else:
        return json_string


def debug_logging(message, dump_message=None):
    """
    print debug_logging

    :param message:
    :param dump_message:
    :return:

    Example:

        .. code-block:: python

            from pawnlib import output

            output.debug_logging("message")

            [2022-07-25 16:35:15.105][DBG][/Users/jinwoo/work/python_prj/pawnlib/examples/asyncio/./run_async.py main(33)] : message

    """
    stack = traceback.extract_stack()
    filename, code_line, func_name, text = stack[-2]

    def_msg = f"[{date_utils.todaydate('log')}][DBG][{filename} {func_name}({code_line})]"
    kvPrint(def_msg, message)
    if dump_message:
        dump(dump_message)


def print_progress_bar(iteration, total, prefix='', suffix='', decimals=1, bar_length=100, overlay=True):
    """
    Print progress bar

    :param iteration:
    :param total:
    :param prefix:
    :param suffix:
    :param decimals:
    :param bar_length:
    :param overlay:
    :return:

    Example:

        .. code-block:: python

            from pawnlib import output

            for i in range(1, 100):
                time.sleep(0.05)
                output.print_progress_bar(i, total=100, prefix="start", suffix="suffix")

            # >> start |\#\#\#\#\#\#\#\| 100.0% suffix

    """
    iteration = iteration + 1
    format_str = "{0:." + str(decimals) + "f}"
    percent = format_str.format(100 * (iteration / float(total)))
    filled_length = int(round(bar_length * iteration / float(total)))
    bar = '#' * filled_length + '-' * (bar_length - filled_length)

    if overlay:
        sys.stdout.write("\033[F")  # back to previous line
        sys.stdout.write("\033[K")  # clear line
    sys.stdout.write('%s |%s| %s%s %s \n' %
                     (prefix, bar, percent, '%', suffix)),

    if iteration == total:
        sys.stdout.write('\n')


def _patched_make_iterencode(markers, _default, _encoder, _indent, _floatstr,
                             _key_separator, _item_separator, _sort_keys, _skipkeys, _one_shot,
                             ## HACK: hand-optimized bytecode; turn globals into locals
                             ValueError=ValueError,
                             dict=dict,
                             float=float,
                             id=id,
                             int=int,
                             isinstance=isinstance,
                             list=list,
                             str=str,
                             tuple=tuple,
                             _intstr=int.__repr__,
                             ):
    if _indent is not None and not isinstance(_indent, str):
        _indent = ' ' * _indent

    def _iterencode_list(lst, _current_indent_level):
        if not lst:
            yield '[]'
            return
        if markers is not None:
            markerid = id(lst)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = lst
        buf = '['
        if _indent is not None:
            _current_indent_level += 1
            newline_indent = '\n' + _indent * _current_indent_level
            # separator = _item_separator + newline_indent
            separator = ", "
            buf += newline_indent

        else:
            newline_indent = None
            separator = _item_separator
        first = True
        for value in lst:
            if first:
                first = False
            else:
                buf = separator
            if isinstance(value, str):
                yield buf + _encoder(value)
            elif value is None:
                yield buf + 'null'
            elif value is True:
                yield buf + 'true'
            elif value is False:
                yield buf + 'false'
            elif isinstance(value, int):
                # Subclasses of int/float may override __repr__, but we still
                # want to encode them as integers/floats in JSON. One example
                # within the standard library is IntEnum.
                yield buf + _intstr(value)
            elif isinstance(value, float):
                # see comment above for int
                yield buf + _floatstr(value)
            else:
                yield buf
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):

                    chunks = _iterencode_dict(value, _current_indent_level)

                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + _indent * _current_indent_level
        yield ']'
        if markers is not None:
            del markers[markerid]

    def _iterencode_dict(dct, _current_indent_level):
        if not dct:
            yield '{}'
            return
        if markers is not None:
            markerid = id(dct)
            if markerid in markers:
                raise ValueError("Circular reference detected")
            markers[markerid] = dct
        yield '{'
        if _indent is not None:
            _current_indent_level += 1
            newline_indent = '\n' + _indent * _current_indent_level
            item_separator = _item_separator + newline_indent
            yield newline_indent
        else:
            newline_indent = None
            item_separator = _item_separator
        first = True
        if _sort_keys:
            items = sorted(dct.items())
        else:
            items = dct.items()
        for key, value in items:
            if isinstance(key, str):
                pass
            # JavaScript is weakly typed for these, so it makes sense to
            # also allow them.  Many encoders seem to do something like this.
            elif isinstance(key, float):
                # see comment for int/float in _make_iterencode
                key = _floatstr(key)
            elif key is True:
                key = 'true'
            elif key is False:
                key = 'false'
            elif key is None:
                key = 'null'
            elif isinstance(key, int):
                # see comment for int/float in _make_iterencode
                key = _intstr(key)
            elif _skipkeys:
                continue
            else:
                raise TypeError(f'keys must be str, int, float, bool or None, '
                                f'not {key.__class__.__name__}')
            if first:
                first = False
            else:
                yield item_separator
            yield _encoder(key)
            yield _key_separator
            if isinstance(value, str):
                yield _encoder(value)
            elif value is None:
                yield 'null'
            elif value is True:
                yield 'true'
            elif value is False:
                yield 'false'
            elif isinstance(value, int):
                # see comment for int/float in _make_iterencode
                yield _intstr(value)
            elif isinstance(value, float):
                # see comment for int/float in _make_iterencode
                yield _floatstr(value)
            else:
                if isinstance(value, (list, tuple)):
                    chunks = _iterencode_list(value, _current_indent_level)
                elif isinstance(value, dict):
                    chunks = _iterencode_dict(value, _current_indent_level)
                else:
                    chunks = _iterencode(value, _current_indent_level)
                yield from chunks
        if newline_indent is not None:
            _current_indent_level -= 1
            yield '\n' + _indent * _current_indent_level
        yield '}'
        if markers is not None:
            del markers[markerid]

    def _iterencode(o, _current_indent_level):
        if isinstance(o, str):
            yield _encoder(o)
        elif o is None:
            yield 'null'
        elif o is True:
            yield 'true'
        elif o is False:
            yield 'false'
        elif isinstance(o, int):
            # see comment for int/float in _make_iterencode
            yield _intstr(o)
        elif isinstance(o, float):
            # see comment for int/float in _make_iterencode
            yield _floatstr(o)
        elif isinstance(o, (list, tuple)):
            yield from _iterencode_list(o, _current_indent_level)

        elif isinstance(o, dict):
            yield from _iterencode_dict(o, _current_indent_level)
        else:
            if markers is not None:
                markerid = id(o)
                if markerid in markers:
                    raise ValueError("Circular reference detected")
                markers[markerid] = o
            o = _default(o)
            yield from _iterencode(o, _current_indent_level)
            if markers is not None:
                del markers[markerid]

    return _iterencode


def json_compact_dumps(data, indent=4, monkey_patch=True):
    if monkey_patch:
        json.encoder._make_iterencode = _patched_make_iterencode
    return json.dumps(data, indent=indent)


class NoIndent(object):
    """ Value wrapper. """

    def __init__(self, value):
        self.value = value


class NoListIndentEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        # Save copy of any keyword argument values needed for use here.
        super(NoListIndentEncoder, self).__init__(*args, **kwargs)

    def iterencode(self, o, _one_shot=False):
        list_lvl = 0
        for s in super(NoListIndentEncoder, self).iterencode(o, _one_shot=_one_shot):
            if s.startswith('['):
                list_lvl += 1
                s = s.replace('\n', '').rstrip()
                s = s.replace(' ', '')
            elif 0 < list_lvl:
                s = s.replace('\n', '').rstrip()
                s = s.replace(' ', '')
                if s and s[-1] == ',':
                    s = s[:-1] + self.item_separator
                elif s and s[-1] == ':':
                    s = s[:-1] + self.key_separator
            if s.endswith(']'):
                list_lvl -= 1
            s = s.replace(",", ", ")
            yield s


class ProgressTime(Progress):
    def __init__(self, **kwargs):
        if kwargs.get('transient', "__NOT_DEFINED__") == "__NOT_DEFINED__":
            kwargs['transient'] = True

        super().__init__(
            SpinnerColumn(),
            *Progress.get_default_columns(),
            TimeElapsedColumn(),
            console=pawn.console,
            **kwargs
        )


def print_syntax(data, name="json", indent=4, style="material", oneline_list=True, line_indent='', rich=True, **kwargs):
    """
    Print the syntax of the data.

    :param data: The data to print.
    :param name: The name of the syntax. Default is "json".
    :param indent: The number of spaces for indentation. Default is 4.
    :param style: The style of the syntax. Default is "material".
    :param oneline_list: Whether to print the list in one line. Default is True.
    :param line_indent: The indentation for each line. Default is ''.
    :param rich: Whether to use rich for printing. Default is True.
    :param kwargs: Other keyword arguments.

    Example:

        .. code-block:: python

            data = {"name": "John", "age": 30, "city": "New York"}
            print_syntax(data)
            # >> {
            # >>     "name": "John",
            # >>     "age": 30,
            # >>     "city": "New York"
            # >> }

            print_syntax(data, name="xml", style="monokai", rich=False)
            # >> <name>John</name>
            # >> <age>30</age>
            # >> <city>New York</city>

    """
    if rich:
        _syntax = syntax_highlight(data, name=name, style=style, rich=True, **kwargs)
        rprint(_syntax)
    else:
        print(syntax_highlight(data, name, indent, style, oneline_list, line_indent))



def syntax_highlight(data, name="json", indent=4, style="material", oneline_list=True, line_indent='', rich=False, **kwargs):
    """
    Syntax highlighting function

    :param data:
    :param name:
    :param indent:
    :param style:
    :param oneline_list:
    :param line_indent:
    :param rich:
    :return:

    Example:

        .. code-block:: python

            from pawnlib import output

            print(output.syntax_highlight("<html><head><meta name='viewport' content='width'>", "html", style=style))

    """
    # styles available as of pygments 2.8.1.
    # ['default', 'emacs', 'friendly', 'colorful', 'autumn', 'murphy', 'manni',
    # 'material', 'monokai', 'perldoc', 'pastie', 'borland', 'trac', 'native',
    # 'fruity', 'bw', 'vim', 'vs', 'tango', 'rrt', 'xcode', 'igor', 'paraiso-light',
    # 'paraiso-dark', 'lovelace', 'algol', 'algol_nu', 'arduino', 'rainbow_dash',
    # 'abap', 'solarized-dark', 'solarized-light', 'sas', 'stata', 'stata-light',
    # 'stata-dark', 'inkpot', 'zenburn']
    if name == "json" and isinstance(data, (dict, list)):
        data = data_clean(data)
        code_data = json_compact_dumps(data, indent=indent, monkey_patch=oneline_list)
    elif data:
        code_data = data
    else:
        code_data = ""

    if line_indent:
        code_data = textwrap.indent(code_data, line_indent)

    if rich:
        return Syntax(code_data, name, theme=style, **kwargs)
    else:
        return highlight(
            code=code_data,
            lexer=get_lexer_by_name(name),
            formatter=Terminal256Formatter(style=style))


def print_here():
    from inspect import currentframe, getframeinfo
    frame_info = getframeinfo(currentframe().f_back)
    filename = frame_info.filename.split('/')[-1]
    line_number = frame_info.lineno
    # loc_str = '%s:%d' % (filename, line_number)
    location_str = f"{filename}:{line_number}"
    print(location_str)


def print_frames(frame_list):
    module_frame_index = [i for i, f in enumerate(frame_list) if f.function == '<module>'][0]
    for i in range(module_frame_index):
        d = frame_list[i][0].f_locals
        local_vars = {x: d[x] for x in d}
        print("  [Frame {} '{}': {}]".format(module_frame_index - i, frame_list[i].function, local_vars))
    print("  [Frame '<module>']\n")


def get_debug_here_info():
    """
    Get debug information from the previous frame.

    This function uses the inspect module to get information about the previous frame, including the filename, line number, function name, lines of context, and index.

    Returns:
        dict: A dictionary containing the debug information.
    """
    filename, line_number, function_name, ln, index = inspect.getframeinfo(inspect.currentframe().f_back.f_back)
    return {
        "filename": filename,
        "line_number": line_number,
        "function_name": function_name,
        "ln": ln,
        "index": index
    }


def get_variable_name_list(var=None):
    """
    Retrieve the name of var.

    :param var: variable to get name from
    :return: name of var

    Example:

        .. code-block:: python

            a = 5
            get_variable_name_list(a)
            # >> ['a']

    """
    callers_local_vars = inspect.currentframe().f_back.f_locals.items()
    return [var_name for var_name, var_val in callers_local_vars if var_val is var]


def get_variable_name(var=None):
    """
    Retrieve the name of the variable.

    :param var: variable to get the name from, defaults to None
    :type var: Any, optional
    :return: name of the variable
    :rtype: str

    Example:

        .. code-block:: python

            a = 10
            get_variable_name(a)
            # >> 'a'

    """
    stacks = inspect.stack()
    try:
        func = stacks[0].function
        code = stacks[1].code_context[0]
        s = code.index(func)
        s = code.index("(", s + len(func)) + 1
        e = code.index(")", s)
        return code[s:e].strip()
    except:
        return ""


def dict_clean(data):
    """
    Clean the dictionary data.

    This function iterates over the items in the dictionary. If the value is an instance of CaseInsensitiveDict, it converts it to a regular dictionary. If the value is None, it converts it to an empty string.

    :param data: The dictionary to clean.
    :type data: dict
    :return: The cleaned dictionary.
    :rtype: dict

    Example:

        .. code-block:: python

            data = {"key1": "value1", "key2": None, "key3": CaseInsensitiveDict({"subkey": "subvalue"})}
            dict_clean(data)
            # >> {"key1": "value1", "key2": "", "key3": {"subkey": "subvalue"}}

    """
    result = {}
    for key, value in data.items():
        if isinstance(value, CaseInsensitiveDict):
            value = dict(value)
        elif value is None:
            value = ''
        result[key] = value
    return result


def list_clean(data):
    """
    Clean the list data.

    This function iterates over the items in the list. If the value is None, it converts it to an empty string.

    :param data: The list to clean.
    :type data: list
    :return: The cleaned list.
    :rtype: list

    Example:

        .. code-block:: python

            data = [None, 'hello', None, 'world']
            list_clean(data)
            # >> ['', 'hello', '', 'world']

    """
    result = []
    for value in data:
        if value is None:
            value = ''
        result.append(value)
    return result


def data_clean(data):
    """
    Clean the data.

    This function checks the type of the data. If the data is a dictionary, it cleans it using the dict_clean function. If the data is a list, it cleans it using the list_clean function.

    :param data: The data to clean.
    :type data: Any
    :return: The cleaned data.
    :rtype: Any

    Example:

        .. code-block:: python

            data = {'name': ' John ', 'age': ' 25 '}
            clean_data = data_clean(data)
            # >> {'name': 'John', 'age': '25'}

            data = [' John ', ' 25 ']
            clean_data = data_clean(data)
            # >> ['John', '25']

    """
    if isinstance(data, dict):
        return dict(dict_clean(data))
    elif isinstance(data, list):
        return list(list_clean(data))
    return data


def count_nested_dict_len(d):
    """
    Count the length of a nested dictionary.

    This function counts the number of items in a dictionary. If a value in the dictionary is also a dictionary, it recursively counts the number of items in that dictionary as well.

    Args:
        d (dict): The dictionary to count the length of.

    Returns:
        int: The length of the dictionary.
    """

    length = len(d)
    for key, value in d.items():
        if isinstance(value, dict):
            length += count_nested_dict_len(value)
    return length


def get_var_name():
    """
    Get the variable name from the call frame.

    This function uses the sys._getframe() function to get the call frame, and then uses the executing library to get the variable name from the call frame.

    Returns:
        str: The variable name if it can be found, otherwise an empty string.
    """

    try:
        call_frame = sys._getframe(2)
        source = executing.Source.for_frame(call_frame)
        ex = source.executing(call_frame)
        func_ast = ex.node
        for ast in func_ast.args:
            if getattr(ast, 'id', ''):
                return ast.id
    except Exception:
        return ""


def get_data_length(data):
    """
    Get the length of the data.

    This function tries to get the length of the data. If the data is a dictionary, it gets the nested dictionary length. If the data is not a dictionary, it gets the length of the data.

    Args:
        data (Any): The data to get the length of.

    Returns:
        str: A string representing the length of the data. If the length cannot be found, it returns an empty string.
    """

    try:
        if isinstance(data, dict):
            return f"nested_dict_len={count_nested_dict_len(data)}"
        else:
            return f"len={len(data)}"
    except Exception:
        return ""


def print_var(data=None, title='', **kwargs):
    """
    Print the variable.

    This function prints the variable with its name, type, and length. It also prints the data if it is a dictionary or a list.

    Args:
        data (Any, optional): The data to print. Defaults to None.
        title (str, optional): The title to print. Defaults to ''.
        **kwargs: Arbitrary keyword arguments.

    Keyword Args:
        line_indent (str): The line indent to use when printing the data. Defaults to '      '.
    """

    kwargs.setdefault('line_indent', '      ')
    var_name = get_var_name()
    data_length = get_data_length(data)
    _title = f"[yellow bold]{title}[/yellow bold]" if title else ""

    pawn.console.log(f"🎁 [[blue bold]{var_name}[/blue bold]] {_title}"
                     f"\t[italic] ({type(data).__name__}), {data_length}", _stack_offset=2)
    if isinstance(data, dict) or isinstance(data, list):
        print(syntax_highlight(data, **kwargs))
    else:
        pawn.console.print(f"\t[white bold]{data}\n")


def print_aligned_text(left_text, right_text, filler='.'):
    """
    Print left and right text aligned with filler in between.

    :param left_text: The text to be aligned on the left side.
    :param right_text: The text to be aligned on the right side.
    :param filler: The character used to fill the space between left and right text. Default is '.'.

    Example:

        .. code-block:: python

            print_aligned_text("Left Text", "Right Text")
            # Left Text...........................................Right Text

            print_aligned_text("Chapter 1", "Page 10", filler='*')
            # Chapter 1********************************************Page 10

    """
    full_text = align_text(left_text, right_text, filler)
    pawn.console.print(full_text)


def align_text(left_text: str = '', right_text: str = '', filler: str = '.', offset: int = 2):
    """
    Aligns text to the left and right with a filler in between.

    :param left_text: The text to align to the left.
    :param right_text: The text to align to the right.
    :param filler: The character to use as filler between the left and right text.
    :param offset: The number of spaces to offset the right text from the right edge.

    Example:

        .. code-block:: python

            align_text('Hello', 'World', filler='-', offset=3)
            # >> 'Hello ---------------------------- World'

            align_text('Left', 'Right', filler='*', offset=5)
            # >> 'Left ************************** Right'

    """
    cleaned_left_text = remove_ascii_color_codes(left_text)
    cleaned_right_text = remove_ascii_color_codes(right_text)

    cleaned_left_text = remove_tags(cleaned_left_text, case_sensitive='lower')
    cleaned_right_text = remove_tags(cleaned_right_text, case_sensitive='lower')

    padding_length = pawn.console.width - len(cleaned_left_text) - len(cleaned_right_text) - offset
    padding = filler * padding_length
    full_text = f"{left_text} {padding} {right_text}"
    return full_text


@contextmanager
def disable_exception_traceback():
    """
    All traceback information is suppressed and only the exception type and value are printed
    """
    default_value = getattr(sys, "tracebacklimit", 1000)  # `1000` is a Python's default value
    sys.tracebacklimit = 0
    yield
    sys.tracebacklimit = default_value  # revert changes
    # try:
    #     sys.tracebacklimit = 0
    #     yield
    # finally:
    #     sys.tracebacklimit = default_value  # revert changes


class NoTraceBackException(Exception):
    def __init__(self, msg):
        try:
            line_no = sys.exc_info()[-1].tb_lineno
            filename = sys.exc_info()[-1].tb_filename
        except AttributeError:
            previous_frame = inspect.currentframe().f_back
            line_no = inspect.currentframe().f_back.f_lineno
            (filename, line_number,
             function_name, ln, index) = inspect.getframeinfo(previous_frame)
        # self.args = "<{0.__name__}> ({2} line {2}): \n {3}".format(type(self), filename, line_no, msg),
        # self.args = "{0}<{1.__name__}>{2} ({3} line {4}): \n {5}".format(bcolors.FAIL, type(self), bcolors.ENDC, filename, line_no, msg),
        # self.args = "<{0.__name__}>({1} line {2}): \n {3}".format(type(self), filename, line_no, msg),
        self.args = "<{0.__name__}> {1}".format(type(self), msg),
        # ex_type, ex_value, traceback = sys.exc_info()
        raise Exception(self)



# -*- coding: utf-8 -*-
import sys
import json
import argparse
from pawnlib.typing import generate_number_list, Counter
from pawnlib.config.globalconfig import ConfigSectionMap, pawnlib_config as pawn
from pawnlib.output import open_file, check_file_overwrite, write_file, is_file, dump
from pawnlib.asyncio import AsyncTasks
from pawnlib.typing import str2bool, flatten_list
from pawnlib.docker import AsyncDocker, run_container, run_dyn_container
from InquirerPy import prompt

from rich.prompt import Confirm, FloatPrompt, Prompt, PromptBase
from rich.syntax import Syntax

from jinja2 import Template
from pathlib import Path
from importlib.machinery import SourceFileLoader
import os
import configparser


EXEC_PATH = os.getcwd()


def get_parser():
    parser = argparse.ArgumentParser(
        description='Command Line Interface for docker',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument(
        'command',
        # choices=['init', 'run', 'create', 'create_or_replace', 'delete', 'start', 'stop'],
        choices=['init', 'run', 'delete', 'start', 'stop', "ls"],
        help='')
    parser.add_argument('-s', '--unixsocket', metavar='unixsocket', help='unix domain socket path (default: /var/run/docker.sock)',
                        default="/var/run/docker.sock")
    parser.add_argument('-d', '--debug', action='store_true', help='debug mode. ', default=False)
    parser.add_argument('-c', '--count', metavar="container count", type=int, help='container count (default: 30)', default=30)
    parser.add_argument('--max-at-once', metavar="max_at_once count", type=int, help='max_at_once count (default: 30)', default=30)
    parser.add_argument('--max-per-second', metavar="max_per_second count", type=int, help='max_per_second count (default: 50)', default=50)
    parser.add_argument('-n', '--name', metavar="container name", type=str, help='container prefix name ', default=None)
    parser.add_argument('-i', '--image', metavar="image name", type=str, help='docker image name ', default=None)

    # parser.add_argument('-f', '--file', type=argparse.FileType('r'), help="container source code")
    parser.add_argument('-f', '--file', type=str, help="import the container source code", default=None)
    parser.add_argument('--force', action='store_true', help="run force mode without confirm", default=False)

    parser.add_argument('--config', type=str, help="import the config file", default=f'{EXEC_PATH}/config.ini')
    # parser.add_argument('-t', '--target', type=str, help="target", default=None)
    parser.add_argument('-t', '--target', nargs='+', help="target", default=None)
    parser.add_argument('-log', '--show-container-log', type=str2bool, help="show last container log (default: False)", default=False)

    return parser


def initialize(args):
    pawn.set(PAWN_CONFIG_FILE=args.config)
    pawn.console.debug(f"os.cwd => {EXEC_PATH}")
    pwn = pawn.conf()
    pawn_config = pawn.to_dict().get('PAWN_CONFIG')

    default_image = {
        "docker_planet": "jinwoo/planet",
        "docker_echo": "jmalloc/echo-server"
    }
    if args.command == "init":
        pawn.console.log("Generate the sample files for docker container")
        template_dir = f"{os.path.dirname(__file__)}/templates"

        template_question = [
            {
                'type': 'list',
                'name': 'category',
                'message': 'What do you want to template?',
                'choices': [tmpl_file.replace(".tmpl", "") for tmpl_file in os.listdir(template_dir) if tmpl_file.endswith('.tmpl')],
            },
            {
                'type': 'input',
                'name': 'target_list',
                'message': 'What\'s the target(access_code) list?',
                # 'default': "[]"
                'default': lambda x: f'Counter(start=10000, stop=10000+args.count, convert_func=str)' if "docker_echo" in x['category'] else "[]"
            },
            {
                'type': 'input',
                'name': 'filename',
                'message': 'What\'s the filename?',
                'default': lambda x: x['category']
            },
            {
                'type': 'input',
                'name': 'image',
                'message': 'What\'s your image name?',
                'default': lambda x: default_image.get(x['category'])
            },
            {
                'type': 'input',
                'name': 'container_name',
                'message': 'What\'s your container name?',
                'default': lambda x: x['category']
            },
        ]

        answers = prompt(template_question)
        template_file = f"{template_dir}/{answers['category']}.tmpl"
        template = open_file(f"{template_file}")

        templated_dict = Template(template).render(
            **answers
        )

        syntax = Syntax(templated_dict, "python")
        pawn.console.rule(answers['category'])
        pawn.console.print(syntax)
        pawn.console.rule("")

        check_file_overwrite(filename=f"{answers['category']}.py")
        write_file(filename=f"{answers['category']}.py", data=templated_dict, permit="660")

        check_file_overwrite(filename=args.config)
        config = configparser.ConfigParser()
        config['default'] = {
            "image": answers['image'],
            "name": answers['container_name'],
            "file": f"{EXEC_PATH}/{answers['filename']}.py",
            "target": answers['target_list'],
        }
        with open(args.config, 'w') as configfile:
            config.write(configfile)

    if pawn_config:
        pawn.console.debug(f"Found config file => '{pwn.PAWN_CONFIG_FILE}'")
        pawn.console.log(f"config=")
        pawn.console.log(pawn_config)

        if not args.image:
            args.image = pawn_config['default']['image']

        if not args.name:
            args.name = pawn_config['default']['name']

        # if not args.target and pawn_config['default'].get('target_list'):
        if not args.target:
            args.target = pawn_config['default'].get('target')

        if not args.file:
            args.file = pawn_config['default'].get('file')

    if args.image and ":" not in args.image:
        args.image = f"{args.image}:latest"

    pawn.set(args=args)

    if args.file and is_file(args.file):
        _file = Path(args.file)
        module_name = _file.stem
        # module_name = args.file.replace(".py", "")
        # pawn.console.debug(f"module_name={module_name}, file={args.file}")
        args.module = SourceFileLoader(module_name, args.file).load_module()
        pawn.console.log(f"[bold]Load module from '{module_name}'")
        pawn.console.debug(f"from {args.module}")
    return args


def _run_async_task(args=None, function=None, target_list=[]):
    if function:
        _function = function
    else:
        _function = args.module.main

    async_tasks = AsyncTasks(args=args, max_at_once=args.max_at_once, max_per_second=args.max_per_second)
    async_tasks.generate_tasks(
        target_list=target_list,
        function=_function,
        **{"args": args, "_config": pawn.conf()}
    )
    async_tasks.run()


def _download_image_if_not_exist(args=None, docker=None):
    if args.image:
        if docker.find_image(image=args.image):
            pawn.console.debug(f"[green][OK] Find Docker Image: {args.image}")
        else:
            pawn.console.log(f"[yellow]Pulling Image: {args.image}")
            docker.pull_image(args.image)
    else:
        pawn.console.log(f"[red] Image argument not found - {args.image}")


def _remove_needless_string(string=None, needless=[]):
    for target_string in needless:
        if string.find(target_string)!= -1:
            string = string.replace(target_string, "")
    return string


def _remove_needless_list(string_list=None, needless=[]):
    result = []
    for string in string_list:
        result.append(_remove_needless_string(string=string, needless=needless))
    return result


def _flatten_string_list(string_list=None):
    result = []
    for string in string_list:
        if string.find("[") != -1 and string.find("]") != -1:
            result.append(eval(f"{string}"))
        else:
            result.append(string)
    return flatten_list(result)


def _convert_string_to_list(string_list, args):
    if isinstance(string_list, str):
        if "[" in string_list and "]" in string_list:
            try:
                string_list = json.loads(string_list)
            except ValueError:
                pawn.console.debug(f"[red] error json load = {string_list}, {type(string_list)}")
        target_list = eval(f"{string_list}")
    elif isinstance(string_list, list):
        target_list = _flatten_string_list(string_list)
    else:
        target_list = string_list
    target_list = _remove_needless_list(target_list, needless=[",", "[", " ", "]"])

    return target_list


def _make_target_list(args=None):
    if args.target:
        target_list = _convert_string_to_list(args.target, args)
    else:
        target_list = _convert_string_to_list(args.module.target_list, args)

    return target_list


def main():
    parser = get_parser()
    args, unknown = parser.parse_known_args()
    args = initialize(args)
    pawn.console.debug(f"args={args}, unknown={unknown}")
    try:
        if args.module:
            pawn.console.debug(f"module={args.module}")
    except Exception as e:
        pawn.console.log(f"[red]Cannot load module - {e}")
        raise ValueError("Did you initialize or Did you installed the \"aiodocker\"? ( pawns docker init )")

    with AsyncDocker(
            client_options=dict(
                url=f"unix://{args.unixsocket}"
            ),
            max_at_once=args.max_at_once,
            max_per_second=args.max_per_second,
            count=args.count,
            container_name=args.name
    ) as docker:
        _download_image_if_not_exist(args, docker)
        target_list = _make_target_list(args)

        if not args.force:
            if args.command == "run":
                pawn.console.log(f"Target={target_list}, Target length={len(target_list)}")
            Confirm.ask(f"Are you sure to {args.command.upper()}?", default=True)

        if args.command == 'run':
            _run_async_task(args=args, target_list=target_list)

        elif args.command == "create":
            async_tasks = AsyncTasks(args=args, max_at_once=args.max_at_once, max_per_second=args.max_per_second)
            async_tasks.generate_tasks(
                target_list=Counter(start=10000, stop=10000+args.count, convert_func=str),
                function=run_container,
                **{"args": args}
            )
            async_tasks.run()
        else:
            docker.control_container(method=args.command, filters={"Names": f"{args.name}.*"}, all=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import argparse
from pawnlib.builder.generator import generate_banner
from pawnlib.__version__ import __version__ as _version
from pawnlib.config import pawnlib_config as pawn
from pawnlib.output import write_json
from pawnlib.resource import server
import pyfiglet

__description__ = 'Command to test the banner.'


def get_parser():
    parser = argparse.ArgumentParser(description='BANNER')
    parser = get_arguments(parser)
    return parser


def get_arguments(parser):
    parser.add_argument("--text",  type=str, help="IP address for retrieving Metadata", default="defaul text")
    return parser


def main():
    banner = generate_banner(
        app_name="banner",
        author="jinwoo",
        description="get the aws metadata",
        font="graffiti",
        version=_version
    )
    print(banner)

    parser = get_parser()
    args, unknown = parser.parse_known_args()

    pawn.console.log(f"args = {args}")
    font_list = pyfiglet.FigletFont.getFonts()

    for font in font_list:
        pawn.console.print(font)
        print(generate_banner(args.text, font=font, version=font))


if __name__ == '__main__':
    main()



from pawnlib.config.globalconfig import pawnlib_config as pawn
from pawnlib.output.file import open_file, write_file, check_file_overwrite
from pawnlib.output.color_print import bcolors, cprint, dump
from jinja2 import Template

from rich.prompt import Confirm, FloatPrompt, Prompt, PromptBase

import os
from pyfiglet import Figlet


def generate_banner(
        app_name: str = "default_app",
        version: str = "{__version}",
        author: str = "Unknown author",
        description: str = "",
        font: str = "big",
        return_type: str = "string",
):
    """

    Generate the banner

    :param app_name:
    :param version:
    :param author:
    :param description:
    :param font: font name   / :refer: http://www.figlet.org/examples.html
    :param return_type: string, list
    :return:

    Example:

    .. code-block:: python

        from pawnlib.builder import generator
        banner = generator.generate_banner(app_name="pawn")
        print(banner)

    """

    result = []
    if return_type == "string":
        enter_string = "\n"
    else:
        enter_string = "\\n"
    result.append(bcolors.WHITE)
    result.append("-" * 50)
    result.append(enter_string)

    ascii_banner = Figlet(font=font)
    for text in ascii_banner.renderText(app_name).split("\n"):
        result.append(text)

    result.append(f" - Description : {description}")
    result.append(f" - Version     : {version}")
    result.append(f" - Author      : {author}")
    result.append(enter_string)
    result.append("-" * 50)
    result.append(bcolors.ENDC)
    if return_type == "string":
        return "\n".join(result)
    elif return_type == "list":
        return result


class AppGenerator:
    """

    :param app_name:
    """
    def __init__(self, app_name="new_app"):

        self.app_name = app_name
        self.cwd = os.getcwd()
        self.template_dir = f"{os.path.dirname(__file__)}/templates"
        self.template_name = "app_with_logging.tmpl"
        self.template = ""
        self.tpl_structure = {}
        self.answers = {}
        print(f"PWD = {self.cwd}")

    def set_user_input(self):
        questions = [
            {
                'type': 'input',
                'name': 'app_name',
                'message': 'What\'s your python3 app name?',
                'default': self.app_name,
            },
            {
                'type': 'input',
                'name': 'author',
                'message': 'What\'s your name?',
                'default': os.getlogin(),
            },
            {
                'type': 'input',
                'name': 'description',
                'message': 'Please explain this script.',
                'default': "This is script",
            },
            {
                'type': 'confirm',
                'name': 'confirm_dir',
                'message': f'Project directory => {self.cwd} ?',
                'default': True,
            },
            {
                'type': 'confirm',
                'name': 'use_logger',
                'message': 'Do you want to logger?',
                'default': True,
            },
            {
                'type': 'confirm',
                'name': 'use_daemon',
                'message': 'Do you want to daemon?',
                'default': False,
            },
        ]
        print("\n")

        for q in questions:
            params = dict(prompt=q.get('message'), default=q.get('default'))
            if q.get("type") == "input":
                self.answers[q.get("name")] = Prompt.ask(**params)
            if q.get("type") == "confirm":
                self.answers[q.get("name")] = Confirm.ask(**params)
        dump(self.answers)
        # dump(self.answers)  # use the answers as input for your app
        # http://www.figlet.org/examples.html

    def load_template(self):
        self.template = open_file(f"{self.template_dir}/{self.template_name}")
        # dump(self.template)

    def get_main_filename(self):
        return f"{self.cwd}/{self.answers.get('app_name', self.app_name)}.py"

    def generate_file_from_template(self):
        self.set_user_input()
        if self.answers.get("app_name", None) is None or self.answers.get("app_name", None) == "":
            self.answers['app_name'] = self.app_name

        if pawn.verbose > 2:
            cprint("answers")
            dump(self.answers)

        self.answers['banner'] = generate_banner(
            app_name=self.answers.get("app_name"),
            author=self.answers.get("author"),
            description=self.answers.get("description"),
            font="rounded",
            return_type="list"
        )
        self.load_template()

        with open(f"{self.template_dir}/{self.template_name}") as app:
            templated_dict = Template(self.template).render(
                **self.answers
            )

            app_filename = f"{self.cwd}/{self.answers['app_name']}.py"
            print(f"app_filename = {app_filename}")
            check_file_overwrite(filename=app_filename)

            if pawn.verbose > 2:
                cprint("templated_dict", "yellow")
                dump(templated_dict)

            write_file(filename=app_filename, data=templated_dict, permit="750")

    def run(self):
        self.generate_file_from_template()
        return f"{self.cwd}/{self.answers['app_name']}.py"



import datetime
import time
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from pawnlib.typing.converter import append_zero
from pawnlib.typing.constants import const


class TimeCalculator:
    """
    A class that converts seconds to a string format.

    :param seconds: The number of seconds to be converted.

    Example:

        .. code-block:: python

            from pawnlib.typing import date_utils
            date_utils.TimeCalculator(1224411)

            # >>  "14 days, 04:06:51"

    """
    def __init__(self, seconds=0):
        self.seconds = seconds
        self._days = 0
        self._hours = 0
        self._minutes = 0
        self._seconds = 0
        self.hhmmss = ""

        self.calculate()

    def calculate(self):
        """
        Calculate the number of days, hours, minutes, and seconds from the given seconds.

        :return: A string representing the number of days, hours, minutes, and seconds.
        """
        seconds = self.seconds
        self._days = int(seconds // const.DAY_IN_SECONDS)
        seconds = int(seconds % const.DAY_IN_SECONDS)
        self._hours = seconds // const.HOUR_IN_SECONDS
        seconds %= const.HOUR_IN_SECONDS
        self._minutes = seconds // const.MINUTE_IN_SECONDS
        seconds %= const.MINUTE_IN_SECONDS
        self._seconds = seconds
        self.hhmmss = "%02i:%02i:%02i" % (self._hours, self._minutes, self._seconds)
        if self._days:
            day_unit = "days"
            if self._days == 1:
                day_unit = "day"
            self.hhmmss = f"{self._days} {day_unit}, {self.hhmmss}"
        return self.hhmmss

    def __str__(self):
        """
        Return the string representation of the calculated time.

        :return: A string representing the number of days, hours, minutes, and seconds.
        """
        return str(self.hhmmss)

    def __repr__(self):
        """
        Return the string representation of the calculated time.

        :return: A string representing the number of days, hours, minutes, and seconds.
        """
        return repr(self.hhmmss)

    def to_strings(self):
        """
        Return the string representation of the calculated time.

        :return: A string representing the number of days, hours, minutes, and seconds.
        """
        return str(self.hhmmss)

    def to_minutes(self):
        """
        Convert the given seconds to minutes.

        :return: The number of minutes.
        """
        return self.seconds // const.MINUTE_IN_SECONDS

    def to_hours(self):
        """
        Convert the given seconds to hours.

        :return: The number of hours.
        """
        return self.seconds // const.HOUR_IN_SECONDS

    def to_days(self):
        """
        Convert the given seconds to days.

        :return: The number of days.
        """
        return self.seconds // const.DAY_IN_SECONDS


def convert_unix_timestamp(date_param):
    """
    Convert a date parameter to a Unix timestamp.

    :param date_param: A date parameter to be converted to a Unix timestamp.
    :type date_param: datetime.datetime or int
    :return: A Unix timestamp.
    :rtype: int

    Example:

        .. code-block:: python

            import datetime

            # Convert a datetime object to a Unix timestamp.
            date = datetime.datetime(2022, 1, 1, 0, 0, 0)
            convert_unix_timestamp(date)
            # >> 1640995200

            # Convert an integer to a Unix timestamp.
            date = 1640995200
            convert_unix_timestamp(date)
            # >> 1640995200
    """
    if isinstance(date_param, datetime.datetime):
        return int(date_param.timestamp())

    if int(date_param) > 1:
        return int(date_param)
    return 0


def get_range_day_of_month(year: int, month: int, return_unix: bool = True):
    """
    This functions will be returned first_day and last_day in parameter.

    :param year:
    :param month:
    :param return_unix:
    :return: (1646060400, 1648738799)

    Example:

        .. code-block:: python

            from pawnlib.typing import date_utils

            date_utils.get_range_day_of_month(year=2022, month=3, return_unix=False)

            # >> ('2022-3-01 00:00:00', '2022-03-31 23:59:59')

    """
    dst_date = datetime.date(year, month, 1)
    next_month = dst_date.replace(day=28) + datetime.timedelta(days=4)
    first_day = f"{year}-{month}-01 00:00:00"
    last_day = f"{next_month - datetime.timedelta(days=next_month.day)} 23:59:59"
    if return_unix:
        first_day = int(time.mktime(datetime.datetime.strptime(first_day, "%Y-%m-%d %H:%M:%S").timetuple()))
        last_day = int(time.mktime(datetime.datetime.strptime(str(last_day), "%Y-%m-%d %H:%M:%S").timetuple()))

    return first_day, last_day


def todaydate(date_type: Literal["file", "time", "time_sec", "hour", "ms", "log", "log_ms", "ms_text", "unix", "ms_unix"] = None) -> str:
    """

    This functions will be returned today date string.

    :param date_type: file, time, time_sec, hour, ms, ms_text, unix, ms_unix
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import date_utils

            date_utils.todaydate("ms_unix")
            # >> '0x5e66be4a93496'

            date_utils.todaydate("log")
            # >> '2022-08-17 17:46:47.281'

    """
    if date_type is None:
        return '%s' % datetime.datetime.now().strftime("%Y%m%d")
    elif date_type == "file":
        return '%s' % datetime.datetime.now().strftime("%Y%m%d_%H%M")
    elif date_type == "time":
        return '%s' % datetime.datetime.now().strftime("%H:%M:%S.%f")[:-3]
    elif date_type == "time_sec":
        return '%s' % datetime.datetime.now().strftime("%H:%M:%S")
    elif date_type == "hour":
        return '%s' % datetime.datetime.now().strftime("%H%M")
    elif date_type == "ms" or date_type == "log" or date_type == "log_ms":
        return '%s' % datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
    elif date_type == "ms_text":
        return '%s' % datetime.datetime.now().strftime("%Y%m%d-%H%M%S%f")[:-3]
    elif date_type == "unix":
        return '%s' % hex(int(datetime.datetime.now().timestamp()))
    elif date_type == "ms_unix":
        return '%s' % hex(int(datetime.datetime.now().timestamp() * 1_000_000))


def format_seconds_to_hhmmss(seconds: int = 0):
    """

    This functions will be returned seconds to hh:mm:ss format

    :param seconds:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import date_utils

            date_utils.format_seconds_to_hhmmss(2323)
            # >> '00:38:43'


    """
    try:
        seconds = int(seconds)
        hours = seconds // const.HOUR_IN_SECONDS
        seconds %= const.HOUR_IN_SECONDS
        minutes = seconds // const.MINUTE_IN_SECONDS
        seconds %= const.MINUTE_IN_SECONDS
        hhmmss = "%02i:%02i:%02i" % (hours, minutes, seconds)
        return hhmmss
    except Exception:
        return seconds


def timestamp_to_string(unix_timestamp: int, str_format='%Y-%m-%d %H:%M:%S'):
    """

    This functions will be returned unix timestamp to hh:mm:ss format

    :param unix_timestamp: unix_timestamp
    :param str_format: string format
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import date_utils

            date_utils.timestamp_to_string(12323232323)

            # >> '2360-07-05 09:05:23'


    """
    ts_length = len(str(unix_timestamp))
    if ts_length == const.SECONDS_DIGITS or ts_length == const.MICRO_SECONDS_DIGITS:
        if ts_length == const.MICRO_SECONDS_DIGITS:
            unix_timestamp = unix_timestamp / 1_000_000
        if unix_timestamp:
            return datetime.datetime.fromtimestamp(unix_timestamp).strftime(str_format)
    raise ValueError('Invalid timestamp')


def second_to_dayhhmm(seconds: int = 0):
    """

    This functions will be returned unix timestamp to days hh:mm:ss format

    :param seconds:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import date_utils

            date_utils.second_to_dayhhmm(2132323)

            # >> '24 days 16:18:43'


    """
    day = int(seconds // const.DAY_IN_SECONDS)
    input_time = int(seconds % const.DAY_IN_SECONDS)
    hour = input_time // const.HOUR_IN_SECONDS
    input_time %= const.HOUR_IN_SECONDS
    minutes = input_time // const.MINUTE_IN_SECONDS
    input_time %= const.MINUTE_IN_SECONDS
    sec = input_time

    return f"{day} days {append_zero(hour)}:{append_zero(minutes)}:{append_zero(sec)}"



import copy
import re
import json
import sys
from pawnlib.config import pawn


def is_json(s) -> bool:
    """
    Check if a string is valid JSON.

    :param s: a string to check if it is valid JSON.
    :return: True if the string is valid JSON, False otherwise.

    Example:

        .. code-block:: python

            check.is_json('{"name": "John", "age": 30, "city": "New York"}')
            # >> True

            check.is_json('{"name": "John", "age": 30, "city": "New York",}')
            # >> False
    """
    try:
        json.loads(s)
    except ValueError:
        return False
    return True


def is_float(s) -> bool:
    """
    Check if a value is float

    :param s: A value to check if it is a float
    :type s: Any
    :return: True if the value is a float, False otherwise
    :rtype: bool

    Example:

        .. code-block:: python

            check.is_float(3.14)
            # >> True

            check.is_float("3.14")
            # >> True

            check.is_float("hello")
            # >> False

    """
    try:
        float(s)
    except (TypeError, ValueError):
        return False
    return True


def is_int(s) -> bool:
    """
    Check if a value is integer.

    :param s: A value to check.
    :type s: Any
    :return: True if the value is an integer, False otherwise.
    :rtype: bool

    Example:

        .. code-block:: python

            check.is_int(1)
            # >> True

            check.is_int(1.0)
            # >> False

            check.is_int("2")
            # >> True

            check.is_int("2.0")
            # >> False

    """
    try:
        int(s)
    except (TypeError, ValueError):
        return False
    return True


def is_hex(s) -> bool:
    """
    Check if a value is hexadecimal

    :param s: string to check
    :return: True if s is hexadecimal, False otherwise

    Example:

        .. code-block:: python

            check.is_hex("1a")
            # >> True

            check.is_hex("g")
            # >> False

    """
    try:
        int(s, 16)
    except TypeError:
        return False
    except ValueError:
        return False
    return True


def is_regex_keyword(keyword: str, value: str) -> bool:
    """
    The is_regex_keyword function takes two strings, a keyword and a value.
    If the keyword starts with / and ends with /, then it is treated as a regex pattern.
    The function checks if the regex pattern is contained within the value string.
    If so, True is returned; otherwise False.

    :param keyword:str: Check if the value:str parameter matches the keyword
    :param value:str: Check if the keyword is in the value
    :return: True if the keyword is a regex and matches

    Example:

        .. code-block:: python

            check.is_regex_keyword("/hello/", "hello world")
            # >> True

            check.is_regex_keyword("(hello)+", "hello world")
            # >> True

            check.is_regex_keyword("hello", "world")
            # >> False
    """
    if len(keyword) <= 0 or len(value) <= 0:
        return False

    if keyword[0] == "/" and keyword[-1] == "/":
        keyword = keyword.replace("/", "")
        if keyword in value:
            return True
    elif keyword[0] == "(" and keyword[-1] == ")":
        if re.findall(keyword, value):
            return True
    else:
        if keyword == value:
            return True


def is_regex_keywords(keywords, value)-> bool:
    """
    Check the value of the keyword regular expression.


    :param keywords:
    :param value:
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing import check

            check.is_regex_keywords(keywords="/sdsd/", value="sdsd")
            # >> True

            check.is_regex_keywords(keywords="/ad/", value="sdsd")
            # >> False

    """
    if not isinstance(keywords, list):
        keywords = [keywords]

    if isinstance(keywords, list):
        for keyword in keywords:
            result = is_regex_keyword(keyword, value)
            if result:
                return True
    return False


def is_valid_ipv4(ip):
    """
    Validates IPv4 addresses.

    :param ip: (str) IPv4 address to validate.
    :return: (bool) True if valid IPv4 address, False otherwise.

    Example:

        .. code-block:: python

            check.is_valid_ipv4("192.168.0.1")
            # >> True

            check.is_valid_ipv4("255.255.255.0")
            # >> True

            check.is_valid_ipv4("300.168.0.1")
            # >> False

    """
    pattern = re.compile(
        r"^((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)\.){3}(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)$",
        re.VERBOSE | re.IGNORECASE
    )
    return pattern.match(ip) is not None


def is_valid_ipv6(ip):
    """
    Validates IPv6 addresses.

    :param ip: A string representing an IPv6 address.
    :return: True if the given string is a valid IPv6 address, False otherwise.

    Example:

        .. code-block:: python

            check.is_valid_ipv6("2001:0db8:85a3:0000:0000:8a2e:0370:7334")
            # >> True

            check.is_valid_ipv6("2001:0db8:85a3::8a2e:0370:7334")
            # >> True

            check.is_valid_ipv6("2001:0db8:85a3:0:0:8a2e:0370:7334:1234")
            # >> False

    """
    pattern = re.compile(r"""
        ^
        \s*                         # Leading whitespace
        (?!.*::.*::)                # Only a single whildcard allowed
        (?:(?!:)|:(?=:))            # Colon iff it would be part of a wildcard
        (?:                         # Repeat 6 times:
            [0-9a-f]{0,4}           #   A group of at most four hexadecimal digits
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
        ){6}                        #
        (?:                         # Either
            [0-9a-f]{0,4}           #   Another group
            (?:(?<=::)|(?<!::):)    #   Colon unless preceeded by wildcard
            [0-9a-f]{0,4}           #   Last group
            (?: (?<=::)             #   Colon iff preceeded by exacly one colon
             |  (?<!:)              #
             |  (?<=:) (?<!::) :    #
             )                      # OR
         |                          #   A v4 address with NO leading zeros 
            (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            (?: \.
                (?:25[0-4]|2[0-4]\d|1\d\d|[1-9]?\d)
            ){3}
        )
        \s*                         # Trailing whitespace
        $
    """, re.VERBOSE | re.IGNORECASE | re.DOTALL)
    return pattern.match(ip) is not None


def is_valid_url(url):
    """
    Check if the given url is valid.

    :param url: (str) url to check
    :return: (bool) True if valid, False otherwise

    Example:

        .. code-block:: python

            check.is_valid_url("google.com")
            # >> True

            check.is_valid_url("http://google.com")
            # >> True

            check.is_valid_url("https://www.google.com/search?q=python")
            # >> True

            check.is_valid_url("ftp://example.com")
            # >> False

    """
    if "http://" not in url and "https://" not in url:
        url = f"http://{url}"

    regex = re.compile(
        r'^https?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+[A-Z]{2,6}\.?|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})' # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)',
        re.IGNORECASE)
    return url is not None and regex.search(url)


def is_valid_private_key(text=None):
    """
    Validates the Private Key

    :param text: A string of private key text.
    :type text: str
    :return: A boolean value indicating whether the private key is valid or not.
    :rtype: bool

    Example:

        .. code-block:: python

            is_valid_private_key("0x1234567890123456789012345678901234567890123456789012345678901234")
            # >> True

            is_valid_private_key("0x12345678901234567890123456789012345678901234567890123456789012345")
            # >> False

    """
    private_length = 64
    if text and is_hex(text):
        if text.startswith("0x"):
            text = text[2:]
        if len(text) == private_length:
            return True
    return False


def is_valid_token_address(text=None, prefix="hx"):
    """
    Validates the token address.

    :param text: A string representing the token address to be validated.
    :param prefix: A string representing the prefix of the token address. Default value is "hx".
    :return: A boolean value indicating whether the token address is valid or not.

    Example:

        .. code-block:: python

            is_valid_token_address("hx1234567890123456789012345678901234567890")
            # >> True

            is_valid_token_address("tx1234567890123456789012345678901234567890")
            # >> False

    """
    if text and prefix \
            and len(text) == 42 \
            and text.startswith(prefix) \
            and is_hex(text[2:]):
        return True
    return False


def is_valid_tx_hash(text=None):
    """
    Validates the txHash

    :param text: A string of txHash text.
    :type text: str
    :return: A boolean value indicating whether the txHash is valid or not.
    :rtype: bool

    Example:

        .. code-block:: python

            is_valid_tx_hash("0x1234567890123456789012345678901234567890123456789012345678901234")
            # >> True

            is_valid_tx_hash("0x12345678901234567890123456789012345678901234567890123456789012345")
            # >> False

    """
    tx_hash_length = 64
    if text and is_hex(text):
        if text.startswith("0x"):
            text = text[2:]
        if len(text) == tx_hash_length:
            return True
    return False


def list_depth(l):
    """
    Returns the depth count of a list.

    :param l: A list.
    :return: An integer representing the depth count of the list.

    Example:

        .. code-block:: python

            list_depth([1, 2, 3])
            # >> 1

            list_depth([1, [2, 3], [4, [5, 6]]])
            # >> 3

    """
    if isinstance(l, list):
        return 1 + max(list_depth(item) for item in l)
    else:
        return 0


def guess_type(s):
    """
    Guess the type of string.

    :param s:
    :return:

    Example:

    .. code-block:: python

        from pawnlib.typing import check

        check.guess_type("True")
        # >> <class 'bool'>

        check.guess_type("2.2")
        # >> <class 'float'>

    """
    s = str(s)
    if s == "":
        return None
    elif re.match(r"^(\d+)\.(\d+)$", s):
        return float
    elif re.match(r"^(\d)+$", s):
        return int
    elif re.match(r"^(true|false)$", s, re.IGNORECASE):
        return bool
    else:
        return str

    # else:
    #     return type(s)


def _str2bool(v) -> bool:
    """
    This function returns boolean type of given string.

    :param v: A string to be converted to boolean type.
    :type v: str
    :return: Boolean value of given string.
    :rtype: bool

    Example:

        .. code-block:: python

            >>> _str2bool('True')
            True
            >>> _str2bool('false')
            False
            >>> _str2bool('1')
            True
            >>> _str2bool(None)
            False
    """
    if v is None:
        return False
    elif type(v) == bool:
        return v
    if isinstance(v, str):
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
    elif v == 1:
        return True

    return False


def return_guess_type(value):
    """
    This function returns the result of :func:`guess_type` and :func:`_strbool`

    :param value: A value to guess the type of.
    :type value: any
    :return: The guessed type of the input value.
    :rtype: any

    Example:

        .. code-block:: python

            return_guess_type("True")
            # >> <class 'bool'>

            return_guess_type("2.2")
            # >> <class 'float'>

    """
    guessed_type = guess_type(value)

    if guessed_type is None or guessed_type == "":
        return value

    if isinstance(guessed_type(), bool):
        return _str2bool(value)
    elif value is not None and value != "":
        return guessed_type(value)
    else:
        return value


def sys_exit(message="", return_code=-1):
    """
    This function executes the sys.exit() method.

    :param message: A message to be printed before exiting. (default="")
    :type message: str
    :param return_code: An exit code to be returned. (default=-1)
    :type return_code: int
    :return: None

    Example:

        .. code-block:: python

            # Example 1: Exit with default return code and message
            sys_exit()

            # Example 2: Exit with custom return code and message
            sys_exit("An error occurred!", 1)

    """
    if message:
        pawn.console.log(f"[red]\[Exit {return_code}] {message}", _stack_offset=2)
    sys.exit(return_code)


def is_include_list(target=None, include_list=[], ignore_case=True):
    """
    Check if target string exists in list.

    :param target: Target string to check.
    :type target: str
    :param include_list: List of strings to check.
    :type include_list: list
    :param ignore_case: If True, ignore case sensitive. Default is True.
    :type ignore_case: bool

    :return: Return True if target string exists in include_list, else False.
    :rtype: bool

    Example:
        .. code-block:: python

            result = is_include_list("hello world", ["hello", "world"])
            # >> True

            result = is_include_list("hello world", ["hello", "world"], ignore_case=False)
            # >> False
    """
    if target and include_list:
        for include_key in include_list:
            if ignore_case and include_key.lower() in target.lower():
                return True
            if include_key in target:
                return True
    return False


def keys_exists(element, *keys):
    """
    Check if **keys** (nested) exists in `element` (dict).
    You don't have to implement it like this.

    [X] if response.get('json') and response['json'].get('result') and response['json']['result'].get('tx_hash'): \n
    [O] if keys_exists(response, 'json', 'result', 'tx_hash'): \n

    :param element: dictionary value
    :param keys: The keys you want to find in the dictionary.
    :return:

    Example:

        .. code-block:: python

            from pawnlib.typing.check import keys_exists

            dict_example = {
                "name": "example",
                "description": {
                    "description_2": "222",
                    "description_3": "333",
                },
                "none_value_key": None,

            }

            keys_exists(dict_example, 'name', 'description')
            # >> True

            keys_exists(dict_example, 'name', 'none_value_key')
            # >> True

            keys_exists(dict_example, 'name', 'none_key')
            # >> False

    """
    not_defined = "__NOT_DEFINED__"
    next_element = ""
    for index, key in enumerate(keys):
        if next_element:
            _element = next_element
        else:
            _element = copy.deepcopy(element)
        if isinstance(_element, dict):
            if _element.get(key, not_defined) != not_defined:
                next_element = _element.pop(key, not_defined)
            else:
                return False
        else:
            return False
    return True


def detect_encoding(byte_data, default_encode="utf8"):
    """
    Detects the encoding of byte data.

    :param byte_data: The byte data to be decoded.
    :type byte_data: bytes
    :param default_encode: The default encoding to be used if no suitable encoding is found. Defaults to "utf8".
    :type default_encode: str
    :return: The detected encoding.
    :rtype: str
    :raises UnicodeDecodeError: If the byte data cannot be decoded using any of the available encodings.

    Examples:

        Detect the encoding of byte data:

        .. code-block:: python

            byte_data = b"\x41\x42\x43"
            detect_encoding(byte_data)
            # Output: 'ascii'

            byte_data = b"\xea\xb0\x80\xeb\x82\x98\xeb\x8b\xa4"
            detect_encoding(byte_data)
            # Output: 'utf8'

            byte_data = b"\xb0\xa1\xb1\xe2\xc6\xae"
            detect_encoding(byte_data)
            # Output: 'euc-kr'
    """

    encodings = ["ascii", "utf8", "euc-kr", "iso2022_jp", "euc_jp", "shift_jis", "cp932", "latin1"]
    for encoding in encodings:
        try:
            byte_data.decode(encoding)
            return encoding
        except UnicodeDecodeError:
            pass
    return default_encode




class TimeConstants:
    MINUTE_IN_SECONDS = 60
    HOUR_IN_SECONDS = 60 * MINUTE_IN_SECONDS
    DAY_IN_SECONDS = 24 * HOUR_IN_SECONDS
    WEEK_IN_SECONDS = 7 * DAY_IN_SECONDS
    MONTH_IN_SECONDS = 30 * DAY_IN_SECONDS
    YEAR_IN_SECONDS = 365 * DAY_IN_SECONDS

    MILLISECOND_IN_SECONDS = 0.001
    MICROSECOND_IN_SECONDS = 0.000001
    NANOSECOND_IN_SECONDS = 0.000000001

    HOUR_IN_MINUTES = 60
    DAY_IN_MINUTES = HOUR_IN_MINUTES * 24
    WEEK_IN_MINUTES = 7 * DAY_IN_MINUTES
    MONTH_IN_MINUTES = 30 * DAY_IN_MINUTES
    YEAR_IN_MINUTES = 365 * DAY_IN_MINUTES


class BooleanConstants:
    TRUE = 1
    FALSE = 0


class NumericConstants:
    TINT = 10 ** 18
    PI = 3.14159
    E = 2.71828
    GOLDEN_RATIO = (1 + 5 ** 0.5) / 2


class AddressConstants:
    ICON_ADDRESS = 42
    ICON_ADDRESS_WITHOUT_PREFIX = 40
    # CHAIN_SCORE_ADDRESS = f"cx{'0'*39}0"
    # GOVERNANCE_ADDRESS = f"cx{'0'*39}1"
    CHAIN_SCORE_ADDRESS = "cx0000000000000000000000000000000000000000"
    GOVERNANCE_ADDRESS = "cx0000000000000000000000000000000000000001"


class TimeStampDigits:
    SECONDS_DIGITS = 10
    MILLI_SECONDS_DIGITS = 13
    MICRO_SECONDS_DIGITS = 16
    NANOSECONDS_DIGITS = 19


class AllConstants(
    TimeConstants,
    BooleanConstants,
    NumericConstants,
    AddressConstants,
    TimeStampDigits,
):
    __slots__ = ()

    def __setattr__(self, name, value):
        raise TypeError("Constants are read-only")


const = AllConstants()



from pawnlib.output import *
from pawnlib.typing.generator import generate_number_list
from pawnlib.asyncio import AsyncTasks, async_partial
import aiodocker
import asyncio
import re
from functools import partial
from devtools import debug
import aiometer
from pawnlib.config import pawnlib_config as pawn


class AsyncDocker:
    def __init__(self, client=None, filters={}, client_options={}, max_at_once=10, max_per_second=10, container_name="", count=0):
        if client:
            self.client = client
        else:
            self.client = aiodocker.Docker(**client_options)
        self.client_options = client_options
        self.image_list = []
        self.container_list = []
        self.loop = None
        self.max_at_once = max_at_once
        self.max_per_second = max_per_second
        self.skip = False
        self.container_name = container_name
        self.count = count

        self.filters = filters
        self.loop_state = {}

        self.default_control_option = {
            "delete": {
                "force": True
            },
            "stats": {
                "stream": False
            }
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pawn.console.debug(f"__exit__ :: before session: {self.client.session.closed} ")
        res = asyncio.run(self.close())
        pawn.console.debug(f"__exit__ :: after session: {res} == session: {self.client.session.closed} ")

    async def close(self):
        await self.client.close()
        if self.loop:
            self.loop.close()

    # @staticmethod
    def run_async_loop(self, function):
        pawn.console.debug(f"Execute function={function.__name__}()")
        self.loop = asyncio.get_event_loop()
        result = self.loop.run_until_complete(function, )
        return result

    def client_decorator(func):
        # debug(f"{func} decorator :: ", f"{skip}")
        async def wrapper(self, *args, **kwargs):
            pawn.console.log(f"execute {func.__name__}() client_decorator() skip={self.skip}")
            pawn.console.log(f"[IN] Decorator args = {args}")

            if self.client:
                pawn.console.log(f"[red] session closed = {self.client.session.closed}")

            if self.skip is False:
                self.client = aiodocker.Docker(**self.client_options)
                debug(self.client.docker_host)

            ret = await func(self, *args, **kwargs)

            if self.skip is False:
                pawn.console.log(f"[red] session closing => {self.client.session.closed}")
                await self.client.close()
                self.skip = False

            return ret

        return wrapper

    # @client_decorator

    async def _async_pull_image(self, image=None):
        try:
            await self.client.images.pull(image)
        except Exception as e:
            pawn.console.log(f"[red][ERROR] {e}")

    async def _async_get_images(self, simple=True):
        # print(f"Simple = {simple}")
        # pawn.console.log(f"---")
        # # self.client = aiodocker.Docker(**self.client_options)
        # pawn.console.log(f"---> {self.client}")
        for image in (await self.client.images.list()):
            # pawn.console.debug(image)
            # tags = image['RepoTags'][0] if image['RepoTags'] else ''
            if isinstance(image['RepoTags'], list):
                tags_list = image['RepoTags']
            else:
                tags_list = [""]

            for tags in tags_list:
                if simple:
                    self.image_list.append({"id": image['Id'].replace("sha256:", "")[:12], "tags": tags})
                else:
                    self.image_list.append(image)

        # await self.client.close()

        return self.image_list

    def pull_image(self, *args, **kwargs):
        return self.run_async_loop(function=self._async_pull_image(*args, **kwargs))

    def get_images(self, *args, **kwargs):
        return self.run_async_loop(function=self._async_get_images(*args, **kwargs))

    def find_image(self, image=None):
        self.get_images()
        for _image in self.image_list:
            if _image.get('tags') == image:
                return True
        return False

    # @client_decorator
    async def _async_get_containers(self, filters={}, **kwargs):
        self.container_list = []
        for container in (await self.client.containers.list(**kwargs)):
            container = self.parse_container_dict(container)
            # debug(f"[get_containers]", container._container['Names'])
            if filters == {} or self.filtering_dict(container._container, filters):
                self.container_list.append(container)

        pawn.console.log(f"[FILTER] condition regex={filters}, matched container={len(self.container_list)}")

        return self.container_list

    def filtering_dict(self, item, filters=None):
        matched = False
        if filters is None:
            filters = self.filters
        if isinstance(item, dict) and isinstance(filters, dict):
            for filter_key, filter_value in filters.items():
                filtering_target = item.get(filter_key, "NNN" * 10)
                # p = re.compile(filter_value)
                # match_regex = re.match(filter_value, filtering_target)
                match_regex = re.fullmatch(fr"{filter_value}", filtering_target)
                # match_regex = re.fullmatch(pattern, filtering_target)
                if match_regex:
                    # cprint(f"[Filter] filter_key={filter_value}, filter_value={filter_value}, "
                    #        f"target={filtering_target}, {match_regex.groups()}")
                    matched = True
        return matched

    def get_containers(self, *args, **kwargs):
        return self.run_async_loop(function=self._async_get_containers(*args, **kwargs))

    def parse_container_dict(self, container):
        # df = pandas.DataFrame(columns=container._container.keys())
        if isinstance(container, aiodocker.containers.DockerContainer):
            # remove the first '/'
            container._container['Names'] = re.sub("^/", "", " ".join(container._container['Names']))
            container._container['Id'] = container._container['Id'][:12]
            container._container['ImageID'] = container._container['ImageID'].replace("sha256:", "")[:12]

            # debug(container._container)
        return container
        # container
        # # df = df.append(container._container.__dict__, ignore_index=True)

    async def delete_container(self, container, container_total, count):
        print(f"[{count}/{container_total}] Delete to {container._id} {container._container['Names']}")
        await container.delete(force=True)

    @staticmethod
    def _callback(function=None, loop_state={}, *args, **kwargs):
        if kwargs:
            pawn.console.log(f"_callback:: {kwargs}")
        res = function(**kwargs)
        container_info = ""
        if loop_state.get('container', None) and isinstance(loop_state['container'], aiodocker.containers.DockerContainer):
            container_info = f" {loop_state['container']._container['Id']}," \
                             f" {loop_state['container']._container['Names']}"

        if loop_state:
            pawn.console.log(f"[{loop_state['count']:>3}/{loop_state['total']}] {function.__name__} the container {container_info}")

            # cprint(f"[{loop_state['count']:>3}/{loop_state['total']}]", "white")
            # f"{function.__name__} the container {loop_state['container']._container['Id']},"
            # f" {loop_state['container']._container['Names']}", "white")
        return res


    async def _generate_echo_containers(self):
        attach_list = []
        for numbering in generate_number_list(start=10000, count=self.count, convert_func=str):
            container_name = f"{self.container_name}_{numbering}"
            options = dict(
                config={
                    'Image': 'jmalloc/echo-server',
                    'Hostname': container_name,
                    'Env': [
                        f"PORT={numbering}"
                    ],
                    "NetworkMode": "host"
                },
                name=container_name,
            )
            attach_list.append(options)
        return attach_list

    async def _print_container_name_in_list(self, containers):
        _count = 0
        pawn.console.rule(f"Container name in list")
        for container in containers:
            container_info = container._container
            pawn.console.log(f"[{_count}] {container_info['Id']}, {container_info['Names']}, {container_info['Status']}")
            _count += 1

    async def _async_control_aiometer_container(self, method, *args, **kwargs):
        if kwargs is None:
            kwargs = {}

        pawn.console.log(f"_async_control_aiometer_container() method={method}, args={args}, "
                         f"kwargs={kwargs}, {self.default_control_option.get(method, {})}")
        results = []
        tasks = []
        count = 0

        if method == "create_or_replace":
            attach_list = self._generate_echo_containers()
        else:
            attach_list = await self._async_get_containers(*args, **kwargs)

        if method == "ls":
            await self._print_container_name_in_list(attach_list)
        else:
            total = len(attach_list)
            for item in attach_list:
                count += 1
                loop_state = {
                    "container": item,
                    "total": total,
                    "count": count
                }
                if isinstance(item, aiodocker.containers.DockerContainer):
                    container = item
                    params = self.default_control_option.get(method, {})
                else:
                    params = item
                    container = self.client.containers

                _function = partial(
                    self._callback,
                    function=getattr(container, method),
                    loop_state=loop_state,
                    # **self.default_control_option.get(method, {})
                    **params
                )

                tasks.append(_function)
            if len(tasks) > 0:
                results = await aiometer.run_all(tasks, max_at_once=self.max_at_once, max_per_second=self.max_per_second)

        return results

    def control_container(self, method=None, *args, **kwargs):
        self.run_async_loop(function=self._async_control_aiometer_container(method=method, *args, **kwargs))

    async def _async_await(self, function=None, method=None, *args, **kwargs):
        container_name = ""
        if isinstance(function, aiodocker.containers.DockerContainer):
            container_name = function._container['Names']

        pawn.console.log(f"_async_await => {function}.{method}() {container_name}")
        if method in ["start"]:
            args = []
            kwargs = {}
        client = aiodocker.Docker()
        function.docker = client
        # func = getattr(function, method)
        res = await getattr(function, method)(*args, **kwargs)
        # await client.close()
        return res

    async def _async_control_container(self, container, container_total=0, count=0, **kwargs):
        # client = aiodocker.Docker().DockerContainer(container)
        client = aiodocker.Docker()
        pawn.console.log(f"[{count}/{container_total}] Delete to {container._id}")
        await client._query_json(f"containers/{container._id}", method="DELETE", params=kwargs)
        await client.close()


async def delete_container(container, container_total=0, count=0):
    print(f"[{count}/{container_total}] Delete to {container._id}")
    await container.delete(force=True)


async def list_things(args, filters={}):
    debug(filters)
    image_list = []
    container_list = []
    print('== Images ==')
    for image in (await args.client.images.list()):
        tags = image['RepoTags'][0] if image['RepoTags'] else ''
        # print(image['Id'], tags)
        image_list.append({"id": image['Id'], "tags": tags})
    print('== Containers ==')
    for container in (await args.client.containers.list()):
        print(f" {container._id}")
        for fk, fv in filters.items():
            filtering_target = container._container.get(fk, "NNN")
            if isinstance(filtering_target, list):
                filtering_target = "".join(filtering_target)
            if fv in filtering_target:
                container_list.append(container)
    return image_list, container_list


async def rm_container(args):
    args.client = aiodocker.Docker()
    image_list, container_list = await list_things(args, filters={"Names": args.name})
    tasks = []
    container_total = len(container_list)
    count = 0
    for container in container_list:
        count += 1
        container_name = "".join(container._container['Names'])
        debug(container._id, container_name)
        # tasks.append(asyncio.ensure_future(container.delete(force=True)))
        tasks.append(partial(delete_container, container, container_total, count))
    if len(tasks) > 0:
        results = await aiometer.run_all(tasks, max_at_once=args.max_at_once, max_per_second=args.max_per_second)
    await args.client.close()


async def run_container(numbering, *args, **kwargs):
    client = aiodocker.Docker()
    args = kwargs.get('args')
    container_name = f"{args.name}_{numbering}"

    if args.__dict__.get('env_key'):
        env_key = args.env_key
    else:
        env_key = "PORT"

    if args.image:
        docker_image = args.image
    else:
        docker_image = "jmalloc/echo-server"

    container_config = {
        'Image': docker_image,
        'Hostname': container_name,
        'Env': [
            f"{env_key}={numbering}"
        ],
        # 'Cmd': ['/bin/ash', '-c', 'echo "hello world"'],
        # "Ports": {
        #     "8080": f"1{numbering}/tcp"
        # },
        "NetworkMode": "host"
    }

    pawn.console.log(f'== Running a "{docker_image}" container => {container_name}, args={args}, container_config={container_config}')

    container = await client.containers.create_or_replace(
        config=container_config,
        name=container_name,
    )
    await container.start()
    logs = await container.log(stdout=True)

    container_name = container.__dict__['_container']['Name'].lstrip("/")
    pawn.console.log(f"container={container_name}, logs={''.join(logs).strip()}")
    # pawn.console.log(''.join(logs))

    await client.close()
    # await container.delete(force=True)
    # await args.client.close()


async def run_dyn_container(config, *args, **kwargs):
    client = aiodocker.Docker()
    pawn.console.debug(f"config={config}")
    pawn.console.debug(f"args={args}")
    pawn.console.debug(f"kwargs={kwargs}")
    container_name = kwargs.get('container_name')

    pawn.console.log(f"[CONTAINER][{pawn.get('count')}] Running container_name={container_name}")

    container = await client.containers.create_or_replace(
        config=config,
        name=container_name,
    )
    await container.start()
    await asyncio.sleep(1)

    if kwargs['args'].show_container_log:
        logs = await container.log(stdout=True, stderr=True, tail=2)
        container_name = container.__dict__['_container']['Name'].lstrip("/")
        pawn.console.log(f"container={container_name}, logs={''.join(logs).strip()}")

    await client.close()
    #


def extract_upper_key_to_env_list(config=None):
    if not config:
        config = pawn.to_dict().get('PAWN_CONFIG')

    result = []
    if config.get('default'):
        for key,value in config['default'].items():
            if key.isupper():
                result.append(f"{key}={value}")
    return result
